import os
import re
import pytest
import pandas as pd
from os.path import join
from battetl.extract import Extractor
from battetl import Constants

BASE_DATA_PATH = os.path.join(os.path.dirname(__file__), 'data')
MACCOR_PATH = os.path.join(BASE_DATA_PATH, 'maccor_cycler_data')
MACCOR_SIMPLE_PATH = os.path.join(MACCOR_PATH, 'simple_data')
MACCOR_FRA_PATH = os.path.join(MACCOR_PATH, 'fra_data')
MACCOR_FASTWAVE_PATH = os.path.join(MACCOR_PATH, 'fastwave_data')
MACCOR_TYPE2_PATH = os.path.join(MACCOR_PATH, 'type2_data')
ARBIN_PATH = os.path.join(BASE_DATA_PATH, 'arbin_cycler_data')
ARBIN_SINGLE_PATH = os.path.join(ARBIN_PATH, 'single_data_file')
ARBIN_MULTIPLE_PATH = os.path.join(ARBIN_PATH, 'multiple_data_files')
ARBIN_CAN_PATH = os.path.join(ARBIN_PATH, 'can_data_files')
ARBIN_SIMULATION_PATH = os.path.join(ARBIN_PATH, 'simulation_data_files')


@pytest.mark.extract
def test_bad_path():
    path = join(
        BASE_DATA_PATH,
        'fake/data/path - 079.txt',
    )

    extractor = Extractor()

    with pytest.raises(TypeError):
        df_extracted = extractor.data_from_files(path)


@pytest.mark.extract
@pytest.mark.maccor
def test_extract_maccor_test_data():
    path = join(MACCOR_SIMPLE_PATH, 'BG_Maccor_Testdata - 079.txt')

    extractor = Extractor()
    extractor.data_from_files([path])
    df_extracted = extractor.raw_test_data

    df_loaded = pd.read_csv(
        path,
        sep='\t',
        header=2,
        skipinitialspace=True,
        index_col=False
    )

    assert (df_extracted.equals(df_loaded))
    assert (extractor.cycler_make == 'maccor')


@pytest.mark.extract
@pytest.mark.maccor
def test_extract_maccor_test_data_type2():
    path = join(MACCOR_TYPE2_PATH, 'BG_Maccor_Type2 - 075.txt')

    extractor = Extractor()
    extractor.data_from_files([path])
    df_extracted = extractor.raw_test_data

    df_loaded = pd.read_csv(
        path,
        sep='\t',
        header=3,
        skipinitialspace=True,
        index_col=False
    )

    assert (df_extracted.equals(df_loaded))
    assert (extractor.cycler_make == 'maccor')


@pytest.mark.extract
@pytest.mark.maccor
def test_extract_maccor_stats_data():
    path = join(MACCOR_SIMPLE_PATH, 'BG_Maccor_TestData - 079 [STATS].txt')

    extractor = Extractor()
    extractor.data_from_files([path])
    df_extracted = extractor.raw_cycle_stats

    df_loaded = pd.read_csv(
        path, sep='\t',
        skiprows=8,
        skipinitialspace=True,
        index_col=False)

    assert (df_extracted.equals(df_loaded))
    assert (extractor.cycler_make == 'maccor')


@pytest.mark.extract
@pytest.mark.maccor
def test_extract_maccor_stats_data_pickle():
    path_pkl = join(MACCOR_SIMPLE_PATH, 'BG_Maccor_TestData - 079.pkl')
    path_txt = join(MACCOR_SIMPLE_PATH, 'BG_Maccor_Testdata - 079.txt')

    extractor = Extractor()
    df_extracted = extractor.from_pickle(path_pkl)

    df_loaded = pd.read_csv(
        path_txt,
        sep='\t',
        header=2,
        skipinitialspace=True,
        index_col=False)

    assert (df_extracted.equals(df_loaded))


@pytest.mark.extract
@pytest.mark.arbin
def test_extract_arbin_single_data_file():
    path = join(ARBIN_SINGLE_PATH,
                'BG_Arbin_TestData_Single_File_Channel_26_Wb_1.CSV')

    extractor = Extractor()
    extractor.data_from_files([path])
    df_extracted = extractor.raw_test_data

    df_loaded = pd.read_csv(path, index_col=False)

    assert (df_extracted.equals(df_loaded))
    assert (extractor.cycler_make == 'arbin')


@pytest.mark.extract
@pytest.mark.arbin
def test_extract_arbin_multiple_data_files():
    paths = [
        join(ARBIN_MULTIPLE_PATH,
             'BG_Arbin_TestData_Multiple_Files_Channel_16_Wb_1.CSV'),
        join(ARBIN_MULTIPLE_PATH,
             'BG_Arbin_TestData_Multiple_Files_Channel_16_Wb_2.CSV'),
        join(ARBIN_MULTIPLE_PATH,
             'BG_Arbin_TestData_Multiple_Files_Channel_16_Wb_3.CSV'),
        join(ARBIN_MULTIPLE_PATH,
             'BG_Arbin_TestData_Multiple_Files_Channel_16_Wb_4.CSV')
    ]

    extractor = Extractor()
    extractor.data_from_files(paths)
    df_extracted = extractor.raw_test_data

    df_1 = pd.read_csv(paths[0], index_col=False)
    df_2 = pd.read_csv(paths[1], index_col=False)
    df_3 = pd.read_csv(paths[2], index_col=False)
    df_4 = pd.read_csv(paths[3], index_col=False)
    df_loaded = pd.concat([df_1, df_2, df_3, df_4], ignore_index=True)

    assert (df_extracted.equals(df_loaded))
    assert (extractor.cycler_make == 'arbin')


@pytest.mark.extract
@pytest.mark.arbin
def test_extract_arbin_globalinfo():
    path = join(ARBIN_SINGLE_PATH,
                'BG_Arbin_TestData_Single_File_Channel_26_GlobalInfo.CSV')

    extractor = Extractor()
    extractor.data_from_files([path])

    assert (
        extractor.raw_test_data_meta_data[0] == {
            'Test Name': 'BG_Characterization_25R_Cell_2',
            'Export Time': '06/01/2022 14:47:10',
            'Serial Number': '213079',
            'Channel': '26',
            'Start DateTime': '05/22/2022 12:42:59',
            'Schedule File Name': 'BG\\BG_25R_Characterization+BG_25R.sdx',
            'Creator': 'admin',
            'Comments': '',
            'Software Version': 'Mits8 PV Build: 202110',
            'Schedule Version': 'Schedule Version 8.00.11',
            'MASS(g)': '0',
            'Specific Capacity(Ah/g)': '0',
            'Capacity(Ah)': '2.5',
            'BarCode': '',
            'Has Aux': '512',
            'Has Specail': '0',
            'Log Aux Data Flag': 'True',
            'Log Special Flag': 'False'
        }
    )


@pytest.mark.extract
@pytest.mark.maccor
@pytest.mark.schedules
def test_extract_maccor_simple_procedure():
    procedure_file_name = 'BG_Maccor_Schedule.000'
    path = join(MACCOR_SIMPLE_PATH, procedure_file_name)

    extractor = Extractor()
    extractor.schedule_from_files([path])
    assert (extractor.schedule['file_name'] == path)
    procedure_extracted = extractor.schedule['schedule']

    assert (extractor.cycler_make == Constants.MAKE_MACCOR)

    schedule_check_dict = {
        'step_types': ['Rest', 'Charge', 'Rest',
                       'AdvCycle', 'Dischrge', 'Rest', 'Charge', 'Rest',
                       'AdvCycle', 'Dischrge', 'Rest', 'Charge', 'Rest',
                       'AdvCycle', 'Dischrge', 'Rest', 'Charge', 'Rest',
                       'AdvCycle', 'Dischrge', 'Rest', 'Charge', 'Rest',
                       'AdvCycle', 'Dischrge', 'Rest', 'Charge', 'Rest',
                       'AdvCycle', 'Dischrge', 'Rest', 'Charge', 'Rest',
                       'AdvCycle', 'Dischrge', 'Rest', 'Charge', 'Rest',
                       'End'],
        'chg_steps': [2, 7, 12, 17, 22, 27, 32, 37],
        'dsg_steps': [5, 10, 15, 20, 25, 30, 35],
        "rest_steps": [1, 3, 6, 8, 11, 13, 16, 18, 21, 23, 26, 28, 31, 33, 36, 38]
    }

    # Check that all step types are correctly defined
    for i, step_type in enumerate(schedule_check_dict['step_types']):
        assert (procedure_extracted['MaccorProcedure']['ProcSteps']
                ['TestStep'][i]['StepType'] == step_type)

    assert (extractor.schedule['steps']['chg']
            == schedule_check_dict['chg_steps'])
    assert (extractor.schedule['steps']['dsg']
            == schedule_check_dict['dsg_steps'])
    assert (extractor.schedule['steps']['rst']
            == schedule_check_dict['rest_steps'])


@pytest.mark.extract
@pytest.mark.maccor
@pytest.mark.schedules
def test_extract_maccor_fra_files():
    procedure_file_name = 'BG_EIS.000'
    fra_file_1 = 'BG_EIS_5mV_BaselineCycle.FRA'
    fra_file_2 = 'BG_EIS_10mV_Pot_DQ1.FRA'
    procedure_path = join(MACCOR_FRA_PATH, procedure_file_name)
    fra_path_1 = join(MACCOR_FRA_PATH, fra_file_1)
    fra_path_2 = join(MACCOR_FRA_PATH, fra_file_2)

    extractor = Extractor()
    extractor.schedule_from_files([procedure_path, fra_path_1, fra_path_2])
    assert (extractor.schedule['file_name'] == procedure_path)
    schedule_extracted = extractor.schedule['schedule']

    fra_dict_1 = schedule_extracted['fra_files'][0][fra_file_1]
    fra_dict_2 = schedule_extracted['fra_files'][1][fra_file_2]

    # Check a sample of values.
    assert (fra_dict_1['General']['fratype'] == '3')
    assert (fra_dict_1['Setup']['range'] == '200 mA')
    assert (fra_dict_1['Sweep']['endfreq'] == '.001')

    assert (fra_dict_2['General']['fratype'] == '3')
    assert (fra_dict_2['Setup']['range'] == '200 mA')
    assert (fra_dict_2['Sweep']['endfreq'] == '1')


@pytest.mark.extract
@pytest.mark.maccor
@pytest.mark.schedules
def test_extract_maccor_fastwave_files():
    procedure_file_name = 'BG_ProfileCycling_MP_CVP_MBC2.000'
    fastwave_file_1 = 'MP_CVP_MBC2.MWF'
    fastwave_file_2 = 'other.MWF'
    procedure_path = join(MACCOR_FASTWAVE_PATH, procedure_file_name)
    fastwave_path_1 = join(MACCOR_FASTWAVE_PATH, fastwave_file_1)
    fastwave_path_2 = join(MACCOR_FASTWAVE_PATH, fastwave_file_2)

    extractor = Extractor()
    extractor.schedule_from_files(
        [procedure_path, fastwave_path_1, fastwave_path_2])
    assert (extractor.schedule['file_name'] == procedure_path)
    schedule_extracted = extractor.schedule['schedule']

    fastwave_string_1 = schedule_extracted['fastwave_files'][0][fastwave_file_1]
    fastwave_string_2 = schedule_extracted['fastwave_files'][1][fastwave_file_2]

    # Check the first and last lines
    assert (re.split('\n', fastwave_string_1)[
            0] == 'C\tI\t6.749\tV\t4.2\t1\tV\t>=\t4.2\tT\t1\tA')
    assert (re.split('\n', fastwave_string_1)
            [-1] == 'C\tI\t1.858\tV\t4.2\t3600\tI\t<=\t0.15\tT\t1\tA')

    assert (re.split('\n', fastwave_string_2)[
            0] == 'C\tI\t6.555\tV\t4.2\t1\tV\t>=\t4.2\tT\t1\tA')
    assert (re.split('\n', fastwave_string_2)
            [-1] == 'C\tI\t1.858\tV\t4.2\t3600\tI\t<=\t0.16\tT\t1\tA')


@pytest.mark.extract
@pytest.mark.arbin
@pytest.mark.schedules
def test_extract_arbin_simple_schedule():
    schedule_file_name = 'BG_25R_Characterization+BG_25R.sdx'
    path = join(ARBIN_SINGLE_PATH, schedule_file_name)

    extractor = Extractor()
    extractor.schedule_from_files([path])
    assert (extractor.schedule['file_name'] == path)
    schedule_extracted = extractor.schedule['schedule']

    schedule_check_dict = {
        'step_types': ['Rest', 'Internal Resistance', 'Current(A)', 'Rest'
                       'Set Variable(s)', 'Internal Resistance', 'Current(A)', 'Rest', 'Current(A)', 'Rest',
                       'Set Variable(s)', 'Internal Resistance', 'CCCV', 'Rest', 'Current(A)', 'Rest',
                       'Set Variable(s)', 'Internal Resistance', 'CCCV', 'Rest', 'Current(A)', 'Rest',
                       'Set Variable(s)', 'Internal Resistance', 'CCCV', 'Rest', 'Current(A)', 'Rest',
                       'Set Variable(s)', 'Internal Resistance', 'CCCV', 'Rest', 'Current(A)', 'Rest',
                       'Set Variable(s)', 'Internal Resistance', 'CCCV', 'Rest', 'Current(A)', 'Rest',
                       'Set Variable(s)', 'Internal Resistance', 'CCCV', 'Rest', 'Current(A)', 'Rest',
                       'Set Variable(s)', 'Current(A)', 'Rest'],
        "chg_steps": [7, 13, 19, 25, 31, 37, 43, 48],
        "dsg_steps": [3, 9, 15, 21, 27, 33, 39, 45],
        "rest_steps": [1, 4, 8, 10, 14, 16, 20, 22, 26, 28, 32, 34, 38, 40, 44, 46, 49]
    }

    # Check that all step types were correctly loaded.
    steps = {key: val for key, val in schedule_extracted['ArbinSchedule']['Schedule'].items()
             if key.startswith('step')}
    for i, step in enumerate(steps):
        assert (steps[step]['m_szStepCtrlType'] ==
                schedule_check_dict['step_types'][i])

    assert (extractor.schedule['steps']['chg']
            == schedule_check_dict['chg_steps'])
    assert (extractor.schedule['steps']['dsg']
            == schedule_check_dict['dsg_steps'])
    assert (extractor.schedule['steps']['rst']
            == schedule_check_dict['rest_steps'])


@pytest.mark.extract
@pytest.mark.arbin
@pytest.mark.schedules
def test_extract_arbin_object_file():
    schedule_file_name = 'BG_25R_Characterization+BG_25R.sdx'
    object_file_name = 'BG_25R.to'
    schedule_path = join(ARBIN_SINGLE_PATH, schedule_file_name)
    object_path = join(ARBIN_SINGLE_PATH, object_file_name)

    extractor = Extractor()
    extractor.schedule_from_files([schedule_path, object_path])
    assert (extractor.schedule['file_name'] == schedule_path)
    schedule_extracted = extractor.schedule['schedule']

    object_dict = schedule_extracted['object_file'][object_file_name]['Content']

    assert (object_dict['m_bautocalcncapacity'] == '0')
    assert (object_dict['m_fmass'] == '0')
    assert (object_dict['m_fmaxcurrentcharge'] == '10')
    assert (object_dict['m_fmaxvoltagecharge'] == '4.2')
    assert (object_dict['m_fminvoltagecharge'] == '2.5')
    assert (object_dict['m_fnorminalcapacitor'] == '0')
    assert (object_dict['m_fnorminalcapacity'] == '2.5')
    assert (object_dict['m_fnorminalir'] == '0')
    assert (object_dict['m_fnorminalvoltage'] == '0')
    assert (object_dict['m_fspecificcapacity'] == '0')
    assert (object_dict['ser'] == '508182206')
    assert (object_dict['ver'] == '27265537')


@pytest.mark.extract
@pytest.mark.arbin
@pytest.mark.schedules
def test_extract_arbin_can_bms_file():
    schedule_file_name = 'BG_MBC_Cycling_ZN_TEST+BG_ATL-GC-SDC-2854A6-020H+model_control_channel_29.sdx'
    can_file_name = 'model_control_channel_29.can'
    schedule_path = join(ARBIN_CAN_PATH, schedule_file_name)
    can_path = join(ARBIN_CAN_PATH, can_file_name)

    extractor = Extractor()
    extractor.schedule_from_files([schedule_path, can_path])
    assert (extractor.schedule['file_name'] == schedule_path)
    schedule_extracted = extractor.schedule['schedule']

    can_dict = schedule_extracted['can_bms_file'][can_file_name]['CANConfig']

    # Just spot check a few because the file is big
    assert (can_dict['Creator'] == 'admin')
    assert (can_dict['CANBaudRate'] == '2')
    assert (can_dict['DataEntryItemList']['DataEntryItem']
            [0]['VariableName'] == 'CAN_MV_RX1')


@pytest.mark.extract
@pytest.mark.arbin
@pytest.mark.schedules
def test_extract_arbin_can_formula_file():
    schedule_file_name = 'BG_MBC_Cycling_ZN_TEST+BG_ATL-GC-SDC-2854A6-020H+model_control_channel_29.sdx'
    can_fm_file_name = 'model_control_channel_29.fm'
    schedule_path = join(ARBIN_CAN_PATH, schedule_file_name)
    formula_path = join(ARBIN_CAN_PATH, can_fm_file_name)

    extractor = Extractor()
    extractor.schedule_from_files([schedule_path, formula_path])
    assert (extractor.schedule['file_name'] == schedule_path)
    schedule_extracted = extractor.schedule['schedule']

    can_fm_dict = schedule_extracted['can_formula_file'][can_fm_file_name]

    # Just spot check a few because the file is big
    assert (can_fm_dict['CAN_AuxDataMaxRow']['datamaxrow'] == '0')
    assert (can_fm_dict['BaudRate']['baudrate'] == '2')
    assert (can_fm_dict['CAN_Formalu1']['lableformalu'] == 'F_A')


@pytest.mark.extract
@pytest.mark.arbin
@pytest.mark.schedules
def test_extract_arbin_mapping_file():
    schedule_file_name = 'BG_MBC_Cycling_ZN_TEST+BG_ATL-GC-SDC-2854A6-020H+model_control_channel_29.sdx'
    batch_file_name = 'ArbinSys.bth'
    schedule_path = join(ARBIN_CAN_PATH, schedule_file_name)
    batch_path = join(ARBIN_CAN_PATH, batch_file_name)

    extractor = Extractor()
    extractor.schedule_from_files([schedule_path, batch_path])
    assert (extractor.schedule['file_name'] == schedule_path)
    schedule_extracted = extractor.schedule['schedule']

    batch_dict = schedule_extracted['mapping_file'][batch_file_name]

    # Just spot check a few because the file is big
    assert (batch_dict['Version Section']
            ['current_version'] == 'Map Version 8.00.02')
    assert (batch_dict['Batch_Test29']['m_uglobalcanbmschanidx'] == '-1')
    assert (batch_dict['Batch']['m_szcreator'] == 'admin')


@pytest.mark.extract
@pytest.mark.arbin
@pytest.mark.schedules
def test_extract_arbin_simulation_files():
    schedule_file_name = 'BG_MBC_eta0p025_25R+BG_25R.sdx'
    sim_file_1 = 'MWT_MBC_eta0p025.txt'
    sim_file_2 = 'sim_file_2.txt'
    schedule_path = join(ARBIN_SIMULATION_PATH, schedule_file_name)
    sim_path_1 = join(ARBIN_SIMULATION_PATH, sim_file_1)
    sim_path_2 = join(ARBIN_SIMULATION_PATH, sim_file_2)

    extractor = Extractor()
    extractor.schedule_from_files([schedule_path, sim_path_1, sim_path_2])
    assert (extractor.schedule['file_name'] == schedule_path)
    schedule_extracted = extractor.schedule['schedule']

    sim_string_1 = schedule_extracted['simulation_files'][0][sim_file_1]
    sim_string_2 = schedule_extracted['simulation_files'][1][sim_file_2]

    # Check the first and last lines in the simulation files
    assert (re.split('\n', sim_string_1)[0] == '1\t7.5')
    assert (re.split('\n', sim_string_1)[-1] == '2059\t0.1')

    assert (re.split('\n', sim_string_2)[0] == '1\t8.5')
    assert (re.split('\n', sim_string_2)[-1] == '2059\t0.2')



import os
import re
import json
import yaml
import dotenv
import logging
import pandas as pd
from collections import OrderedDict
from pydash import get, set_with, unset, merge

from battetl import logger, Constants


class Utils:
    def load_env(env_path: str) -> None:
        """
        Set environment variables from .env file

        Parameters
        ----------
        env_path : str
            Path to .env file
        """
        logger.info(f'Load environment variables from {env_path}')
        dotenv.load_dotenv(env_path, override=True)
        # Set logger level
        if os.getenv('ENV') == 'dev':
            logger.info('Set logger level to DEBUG')
            logger.setLevel(logging.DEBUG)
        elif os.getenv('ENV') in ['prd', 'prod']:
            logger.info('Set logger level to WARNING')
            logger.setLevel(logging.WARNING)
        
    def load_config(config_path: str) -> dict:
        """
        Load config file from path

        Parameters
        ----------
        config_path : str
            Path to config file

        Returns
        -------
        config : dict
            Config file as dictionary
        """
        logger.info(f'Load config file from {config_path}')
        config = None

        if not os.path.exists(config_path):
            raise FileNotFoundError(f'Config file not found at {config_path}')

        try:
            with open(config_path, 'r') as f:
                config = json.load(f)
                logger.info('Loaded config file as json')
                return config
        except json.decoder.JSONDecodeError:
            logger.debug('Config file is not json')

        try:
            with open(config_path, 'r') as f:
                config = yaml.load(f, Loader=yaml.SafeLoader)
                logger.info('Loaded config file as yaml')
                return config
        except yaml.YAMLError:
            logger.debug('Config file is not yaml')

        # If we get here, the config file is not json or yaml
        raise Exception('Config file is not json or yaml')

    def drop_unnamed_columns(df: pd.DataFrame) -> pd.DataFrame:
        """
        The function drops unnamed columns from DataFrame

        Parameters
        ----------
        df : pandas.DataFrame
            DataFrame to drop unnamed columns from.

        Returns
        -------
        df : pandas.DataFrame
            Pandas DataFrame with the unnamed columns dropped.
        """
        logger.info('Drop unnamed columns')
        unnamed_column_list = df.filter(like='Unnamed', axis=1).columns
        df = Utils.drop_columns(df, unnamed_column_list)

        return df

    def drop_columns(df: pd.DataFrame, columns_to_drop: list[str]) -> pd.DataFrame:
        """
        This function drops unnamed columns from the passed DataFrame.

        Parameters
        ----------
        df : pandas.DataFrame
            DataFrame to drop unnamed columns from.
        columns_to_drop : list
            List of strings that give column names to drop

        Returns
        -------
        df : pandas.DataFrame
            Pandas DataFrame with the dropped columns.
        """
        for column in columns_to_drop:
            if column in df.columns:
                logger.info(f'Drop column {column}')
                df = df.drop(columns=column)

        return df

    def drop_empty_rows(df: pd.DataFrame) -> pd.DataFrame:
        """
        The function drops empty rows from DataFrame

        Parameters
        ----------
        df : pandas.DataFrame
            DataFrame to drop empty rows from.

        Returns
        -------
        df : pandas.DataFrame
            Pandas DataFrame with the empty rows dropped.
        """
        all_na_count = (df.isna().all(axis=1)).sum()
        if all_na_count > 0:
            logger.info(f'Drop {all_na_count} empty rows')
            df = df.dropna(how='all')

        return df

    def get_cycle_make(columns: list) -> tuple[str, str]:
        """
        Determine the make and type of cycler.
        Currently supported:
        - Arbin Test Data
        - Arbin Cycle Stats
        - Maccor Test Data
        - Maccor Cycle Stats

        Parameters
        ----------
        data : list[str]
            Original data 

        Returns
        -------
        (str)
            Cycle make
        (str)
            Data type
        """
        columnsSet = Utils.get_lower_strip_set(columns)
        arbinTestDataSet = Utils.get_lower_strip_set(
            Constants.COLUMNS_ARBIN_TEST_DATA_ONLY)
        arbinCycleStatsSet = Utils.get_lower_strip_set(
            Constants.COLUMNS_ARBIN_CYCLE_STATS_ONLY)
        maccorTestDataSet = Utils.get_lower_strip_set(
            Constants.COLUMNS_MACCOR_TEST_DATA_ONLY)
        maccorTestDataSetType2 = Utils.get_lower_strip_set(
            Constants.COLUMNS_MACCOR_TEST_DATA_TYPE2_ONLY)
        maccorTestDataSetCustomer1 = Utils.get_lower_strip_set(
            Constants.COLUMNS_MACCOR_TEST_DATA_CUSTOMER1)
        maccorCycleStatsSet = Utils.get_lower_strip_set(
            Constants.COLUMNS_MACCOR_CYCLE_STATS_ONLY)
        maccorCycleStatsSetCustomer1 = Utils.get_lower_strip_set(
            Constants.COLUMNS_MACCOR_CYCLE_STATS_CUSTOMER1)

        if len(columnsSet & arbinTestDataSet) >= (len(arbinTestDataSet) / 2):
            return Constants.MAKE_ARBIN, Constants.DATA_TYPE_TEST_DATA

        if len(columnsSet & arbinCycleStatsSet) >= (len(arbinCycleStatsSet) / 2):
            return Constants.MAKE_ARBIN, Constants.DATA_TYPE_CYCLE_STATS

        if len(columnsSet & maccorTestDataSet) >= (len(maccorTestDataSet) / 2):
            return Constants.MAKE_MACCOR, Constants.DATA_TYPE_TEST_DATA

        if len(columnsSet & maccorTestDataSetType2) >= (len(maccorTestDataSetType2) / 2):
            return Constants.MAKE_MACCOR, Constants.DATA_TYPE_TEST_DATA

        if len(columnsSet & maccorTestDataSetCustomer1) >= (len(maccorTestDataSetCustomer1) / 2):
            return Constants.MAKE_MACCOR, Constants.DATA_TYPE_TEST_DATA

        if len(columnsSet & maccorCycleStatsSet) >= (len(maccorCycleStatsSet) / 2):
            return Constants.MAKE_MACCOR, Constants.DATA_TYPE_CYCLE_STATS

        if len(columnsSet & maccorCycleStatsSetCustomer1) >= (len(maccorCycleStatsSetCustomer1) / 2):
            return Constants.MAKE_MACCOR, Constants.DATA_TYPE_CYCLE_STATS

        return None, None

    def get_lower_strip_set(data: list[str]) -> set:
        """
        Transform string list to lower case set

        Parameters
        ----------
        data : list[str]
            Original data 

        Returns
        -------
        (set)
            Lower case set data
        """
        return set([d.lower().strip().replace(' ', '').replace('_', '') for d in data])

    def rename_df_columns(df: pd.DataFrame, columnsMapping: dict) -> pd.DataFrame:
        """
        Rename column names to BattETL format

        Parameters
        ----------
        df : pandas.DataFrame
            Original data

        Returns
        -------
        df : pandas.DataFrame
            Renamed data
        """
        logger.info('Rename column names to BattETL format')
        logger.debug(f'Columns before renaming: {df.columns.values}')

        # Normalize mapping data
        mappingTable = {k.lower().strip(): v for k,
                        v in columnsMapping.items()}

        # Normalize DataFrame columns
        df.columns = df.columns.str.lower()
        df.columns = df.columns.str.strip()

        # Rename thermocouple columns
        for col_name in df.columns.values:
            match_arbin = re.search(
                Constants.PREFIX_ARBIN_THERMOCOUPLE+r'(\d)+ \(c\)', col_name)
            match_maccor = re.search(
                Constants.PREFIX_MACCOR_THERMOCOUPLE + r'(\d)', col_name)
            match = (match_arbin or match_maccor)
            if match:
                new_col_name = Constants.TEMPLATE_RENAMED_THERMOCOUPLE.replace(
                    'X', match.group(1))
                df = df.rename(columns={col_name: new_col_name})

        # Rename to BattETL format
        df = df.rename(mappingTable, axis='columns')
        logger.debug(f'Columns after renaming: {df.columns.values}')

        return df

    def convert_to_milli(df: pd.DataFrame) -> pd.DataFrame:
        """
        Convert columns to milli- and rename column name

        Parameters
        ----------
        df : pandas.DataFrame
            Original data 

        Returns
        -------
        df : pandas.DataFrame
            Converted data
        """
        logger.info('Convert data to milli-')
        for column in df.columns:
            if column in Constants.COLUMNS_TO_MILLI:
                logger.debug(f'Converting {column}')
                df[column] = df[column].replace(
                    {',': ''}, regex=True).apply(pd.to_numeric, 1)
                df[column] = df[column] * 1e3
                columnName = Constants.COLUMNS_TO_MILLI[column]
                logger.debug(f'Rename column name to {columnName}')
                df = df.rename({column: columnName}, axis='columns')

        return df

    def sort_dataframe(df: pd.DataFrame, columns: list[str]) -> pd.DataFrame:
        """
        Sort pandas.DataFrame with input columns

        Parameters
        ----------
        df : pandas.DataFrame
            Original data 
        columns : list[str]
            Sort by columns

        Returns
        -------
        df : pandas.DataFrame
            Sorted data
        """
        logger.info(f'Sort DataFrame by {columns}')

        df = df.sort_values(by=columns)
        df = df.reset_index(drop=True)
        logger.debug(f'Sorted rows: {df.shape[0]}')

        return df

    def convert_timedelta_to_seconds(df: pd.DataFrame, column: str) -> pd.DataFrame:
        """
        Convert time delta to seconds

        Parameters
        ----------
        df : pandas.DataFrame
            Original data 
        column : str
            Column to convert

        Returns
        -------
        df : pandas.DataFrame
            Converted data
        """
        logger.info(f'Convert {column} to seconds')

        if column not in df.columns:
            raise NameError(f'Can not find {column}')

        df[column] = round(pd.to_timedelta(df[column]).dt.total_seconds(), 3)

        return df

    def convert_datetime(df: pd.DataFrame, column: str, timezone: str) -> pd.DataFrame:
        """
        Convert datetime to UTC format with time zone

        Parameters
        ----------
        df : pandas.DataFrame
            The input DataFrame
        column : str
            column to convert
        timezone : str
            Time zone strings in the IANA Time Zone Database

        Returns
        -------
        df : pandas.DataFrame
            Converted data
        """
        logger.info(f'Convert {column} to UTC with timezone {timezone}')

        if column not in df.columns:
            raise NameError(f'Can not find {column}')

        df[column] = Utils.__parse_datetime(df, column)
        df[column] = df[column].dt.tz_localize(timezone)
        df[column] = df[column].dt.tz_convert('UTC')

        return df

    def __parse_datetime(df: pd.DataFrame, column: str) -> pd.DataFrame:
        """
        Parse datetime with default format to speed up `pandas.to_datetime`

        This function speeds up the `pandas.to_datetime` process time by
        set the date format instead of traversing each row of data.
        The function first tries our default formats,
        and if it doesn't work, it falls back to the original method.

        Parameters
        ----------
        df : pandas.DataFrame
            The input DataFrame
        column : str
            column to convert

        Returns
        -------
        data : pandas.DataFrame
            Converted data
        """
        format_list = [
            '\t%m/%d/%Y %H:%M:%S.%f',
            '%m/%d/%Y %H:%M:%S.%f',
            '%m/%d/%Y %I:%M:%S %p',
            '%m/%d/%Y %H:%M:%S',
        ]
        for format in format_list:
            try:
                data = pd.to_datetime(df[column], format=format)
                logger.debug(
                    f'Found datetime format "{format}" for column {column}')
                return data
            except:
                pass

        logger.debug(
            f'Can not find format in {format_list}, use default')
        return pd.to_datetime(df[column])

    def convert_to_float(value):
        '''
        Converts value to float if it is a string.

        Parameters
        ----------
        value : str or float
            The value to convert.

        Returns
        -------
        float
        '''
        if isinstance(value, str):
            return float(value.replace(',', ''))
        return value
    
    def validate_file_meta(file_meta: dict) -> bool:
        """
        Validate file_meta
        Parameters
        ----------
        file_meta : dict
            Dictionary containing the meta data for the file.
            For example:
            ```json
            {
                "voltage_mv" : 
                {
                    "column_name":"volt",
                    "scaling_factor":1,
                },
                "current_ma" : 
                {
                    "column_name":"curr",
                    "scaling_factor":1,
                },
            }
            ```
        Returns
        -------
        bool
            True if valid
        """
        # Check file_meta contains all required keys
        for key in Constants.UNSTRUCTURED_DATA_REQUIRED_KEYS:
            if key not in file_meta:
                raise ValueError(
                    f'file_meta does not contain required key: {key}')

        return True


class DashOrderedDict(OrderedDict):
    """
    Pulled from BEEP:
    https://github.com/TRI-AMDD/beep/blob/master/beep/utils/__init__.py

    Nested data structure with pydash enabled
    getters and setters.  Nested values can
    be set using dot notation, e. g.

    >>> dod = DashOrderedDict()
    >>> dod.set('key1.key2', 5)
    >>> print(dod['key1']['key2'])
    >>> 5
    """

    def set(self, string, value):
        set_with(self, string, value, lambda x: OrderedDict())

    def get_path(self, string, default=None):
        return get(self, string, default=default)

    def unset(self, string):
        unset(self, string)

    def merge(self, obj):
        merge(self, obj)

    def __str__(self):
        return "{}:\n{}".format(self.__class__.__name__, json.dumps(self, indent=4))

    def __repr__(self):
        return self.__str__()



import os
import pytest
import pandas as pd
from pathlib import Path

from battetl import Utils

CONFIG_DIR = Path(__file__).parent / 'configs'
ENV_PATH = Path(__file__).parent / '.env.example'


@pytest.mark.utils
def test_utils_convert_datetime():
    # Create a DataFrame
    # 0	2023-01-01 00:00:00
    # 1	2023-01-01 01:00:00
    # 2	2023-01-01 02:00:00
    df = pd.DataFrame({'datetime': pd.date_range(
        "2023-01-01", periods=3, freq="H")})

    # UTC-08:00
    df_1 = Utils.convert_datetime(df.copy(), 'datetime', 'America/Los_Angeles')
    dt_1 = df_1.iloc[0]['datetime'].strftime('%Y-%m-%d %H:%M:%S')

    # UTC-07:00
    df_2 = Utils.convert_datetime(df.copy(), 'datetime', 'America/Phoenix')
    dt_2 = df_2.iloc[0]['datetime'].strftime('%Y-%m-%d %H:%M:%S')

    # UTC+01:00
    df_3 = Utils.convert_datetime(df.copy(), 'datetime', 'Europe/Berlin')
    dt_3 = df_3.iloc[0]['datetime'].strftime('%Y-%m-%d %H:%M:%S')

    # UTC+00:00 to UTC-08:00
    assert (dt_1 == '2023-01-01 08:00:00'), 'UTC-08:00. Should be 2023-01-01 08:00:00'

    # UTC+00:00 to UTC-07:00
    assert (dt_2 == '2023-01-01 07:00:00'), 'UTC-07:00. Should be 2023-01-01 07:00:00'

    # UTC+00:00 to UTC+01:00
    assert (dt_3 == '2022-12-31 23:00:00'), 'UTC+01:00. Should be 2022-12-31 23:00:00'


@pytest.mark.utils
def test_utils_load_env(env_example):
    Utils.load_env(ENV_PATH)
    credentials = {
        'ENV': os.getenv('ENV'),
        'DB_HOSTNAME': os.getenv('DB_HOSTNAME'),
        'DB_PORT': os.getenv('DB_PORT'),
        'DB_TARGET': os.getenv('DB_TARGET'),
        'DB_USERNAME': os.getenv('DB_USERNAME'),
        'DB_PASSWORD': os.getenv('DB_PASSWORD'),
    }
    assert credentials == env_example, 'Env not loaded correctly'


@pytest.mark.utils
class TestUtilsLoadConfig:
    def test_utils_load_json_config(self, cell_config):
        config_path = Path(CONFIG_DIR) / 'config_1.json'

        config = Utils.load_config(config_path)
        assert config == cell_config, 'JSON config not loaded correctly'

    def test_utils_load_yaml_config(self, cell_config):
        config_path = Path(CONFIG_DIR) / 'config_1.yaml'

        config = Utils.load_config(config_path)
        assert config == cell_config, 'YAML config not loaded correctly'

    def test_utils_load_invalid_config_path(self):
        config_path = Path(CONFIG_DIR) / 'invalid_path'

        with pytest.raises(FileNotFoundError):
            config = Utils.load_config(config_path)


@pytest.mark.utils
def test_utils_drop_empty_rows():
    test_df = pd.DataFrame({
        'A': [1.0, 2.0, None],
        'B': [None, None, None],
        'C': ['a', 'b', None],
    })
    assert len(test_df) == 3, 'Test DataFrame should have 3 rows'
    new_df = Utils.drop_empty_rows(test_df)
    assert len(new_df) == 2, 'New DataFrame should have 2 rows'
    assert new_df.equals(pd.DataFrame({
        'A': [1.0, 2.0],
        'B': [None, None],
        'C': ['a', 'b'],
    })), 'New DataFrame should equal expected DataFrame'



import os
import re
import json
import xmltodict
import configparser
import pandas as pd
from collections import OrderedDict

from battetl import logger, Constants, Utils
from battetl.utils import DashOrderedDict


class Extractor:
    def __init__(self):
        """
        An interface to extract battery test data from raw data files. 
        """
        self.raw_test_data_meta_data = []
        self.raw_cycle_stats_meta_data = []
        self.raw_test_data = pd.DataFrame(dtype=object)
        self.raw_cycle_stats = pd.DataFrame(dtype=object)
        self.cycler_make = ''

        self.schedule = {
            'schedule': DashOrderedDict(),
            'file_name': '',
            'steps': {'chg': [], 'dsg': [], 'rst': []}
        }

    def data_from_files(self, paths: list[str], file_meta: dict = None) -> pd.DataFrame:
        """
        Extracts multiple test data files into a single pandas DataFrame.
        If only a single data file exists then it only extracts that data.

        Parameters
        ----------
        paths : list[str]
            Relative or absolute paths to the target data files.
        file_meta : dict, optional
            Dictionary containing the user defined column names for the test data. The default is None.
        Returns
        -------
        pd.DataFrame
            A pandas DataFrame containing the test data.
        """
        if type(paths) != list:
            raise TypeError('Input paths is not list')

        logger.info(f'Total {len(paths)} files')

        for path in paths:
            if file_meta:
                self.__unstructured_data_from_file(
                    path=path,
                    file_meta=file_meta)
            else:
                self.__data_from_file(path=path)

        logger.info('Extract success')

    def schedule_from_files(self, paths: list[str]) -> dict:
        """
        Reads Arbin schedules and associated files or Maccor procedures and associated 
        files from the passed paths list and saves them together in a nested dictionary.

        Parameters
        ----------
        paths : list[str]
            Relative or absolute paths to the Maccor schedule and associated files
        """
        if type(paths) != list:
            raise TypeError('Input paths is not list')

        logger.info(f'Total {len(paths)} files')

        for i, path in enumerate(paths):
            if path.endswith('.000'):
                logger.info("Processing procedure files for Maccor")
                self.schedule['schedule'] = self.__maccor_procedure_from_files(
                    paths)
                break
            elif path.endswith('.sdu') or path.endswith('.sdx'):
                logger.info("Processing schedule files for Arbin")
                self.schedule['schedule'] = self.__arbin_schedule_from_files(
                    paths)

                break
            elif i == (len(paths)-1):
                logger.error(
                    'Unable to identify schedule type from paths: ' + str(paths))

        return self.schedule

    def schedule_from_files(self, paths: list[str]) -> dict:
        """
        Reads Arbin schedules and associated files or Maccor procedures and associated 
        files from the passed paths list and saves them together in a nested dictionary.

        Parameters
        ----------
        paths : list[str]
            Relative or absolute paths to the Maccor schedule and associated files
        """
        if type(paths) != list:
            raise TypeError('Input paths is not list')

        logger.info(f'Total {len(paths)} files')

        for i, path in enumerate(paths):
            if path.endswith('.000'):
                logger.info("Processing procedure files for Maccor")
                self.schedule['schedule'] = self.__maccor_procedure_from_files(
                    paths)
                break
            elif path.endswith('.sdu') or path.endswith('.sdx'):
                logger.info("Processing schedule files for Arbin")
                self.schedule['schedule'] = self.__arbin_schedule_from_files(
                    paths)

                break
            elif i == (len(paths)-1):
                logger.error(
                    'Unable to identify schedule type from paths: ' + str(paths))

        return self.schedule

    def __unstructured_data_from_file(self, path: str, file_meta: dict) -> pd.DataFrame:
        """
        Reads unstructured data from the passed file path and returns it as a pandas DataFrame.
        Parameters
        ----------
        path : str
            Relative or absolute path to the datafile.
        file_meta : dict
            Dictionary containing the meta data for the file.
        Returns
        -------
        df : pandas.DataFrame
            A pandas DataFrame containing the data file.
        """
        logger.info(f'Load file path: {path}')
        logger.info(f'Unstructured data from file. File meta: {file_meta}')
        if not os.path.exists(path):
            raise FileNotFoundError(f'Unable to load file {path}')

        # Check file_meta contains all required keys
        Utils.validate_file_meta(file_meta)

        df = pd.read_csv(path, **file_meta['pandas_read_csv_args'])

        self.raw_test_data = pd.concat(
            [
                self.raw_test_data,
                df
            ],
            ignore_index=True
        )
        logger.debug(
            f'Update raw_test_data. Total rows: {self.raw_test_data.shape[0]}')

        return df

    def __maccor_procedure_from_files(self, paths: list[str]) -> DashOrderedDict:
        """
        Reads the Maccor procedure and associated files from the passed file path list.
        Returns all files a dictionaries in a single nested DashOrderedDict. Supported files to 
        include in paths list are Maccor procedures (.000) (required),  FastWave (*.MWF)  
        and frequency response analyzer (*.FRA) files.

        Parameters
        ----------
        paths : list[str]
            Relative or absolute paths to the Maccor schedule and associated files

        Returns
        -------
        procedure : DashOrderedDict
            The Maccor procedure and associated files saved as a nested dict.
        """
        procedure_dod = DashOrderedDict()
        fra_dicts = []
        fastwave_dicts = []

        if self.cycler_make == '':
            self.cycler_make = Constants.MAKE_MACCOR
        elif self.cycler_make != Constants.MAKE_MACCOR:
            logger.error(
                "Passed schedule does not match previously assigned cycler_make!")
            return {}

        for path in paths:
            if path.endswith('.000'):
                if not self.schedule['file_name']:
                    self.schedule['file_name'] = path
                    logger.info(
                        f'Importing Maccor procedure {path}')
                    procedure = self.__procedure_from_file(path)
                else:
                    logger.error("A schedule file name is already defined!")
            elif path.endswith('.FRA'):
                logger.info(
                    f'Importing FRA file {path}')
                fra_dicts.append(self.__fra_from_file(path))
            elif path.endswith('.MWF'):
                logger.info(
                    f'Importing fastwave file {path}')
                fastwave_dicts.append(self.__fastwave_from_file(path))
            else:
                logger.warning(f'Unknown file type included: {path}')

        try:
            procedure_dod.set('MaccorProcedure', procedure)
        except NameError:
            logger.error(
                "Missing Maccor procedure file from included paths! " + str(paths))
            return {}

        self.schedule['steps'] = self.__steps_from_procedure(procedure)

        if fastwave_dicts:
            procedure_dod.set('fastwave_files', fastwave_dicts)
        if fra_dicts:
            procedure_dod.set('fra_files', fra_dicts)

        return procedure_dod

    def __procedure_from_file(self, path) -> dict:
        """
        Reads the Maccor procedure file (*.000) and returns the contents 
        as a nested dict. This function was pulled from BEEP:
        https://github.com/TRI-AMDD/beep/blob/master/beep/protocol/maccor.py

        Parameters
        ----------
        path : str
            Relative or absolute path to the procedure file.

        Returns
        -------
        procedure : dict
            The Maccor procedure as a nested dictionary.
        """
        with open(path, "rb") as f:
            text = f.read().decode(Constants.MACCOR_PROCEDURE_FILE_ENCODING)
        procedure = xmltodict.parse(
            text, process_namespaces=False, strip_whitespace=True)
        return procedure['MaccorTestProcedure']

    def __fra_from_file(self, path) -> dict:
        """
        Reads the Maccor frequency response analyzer file (*.FRA) and returns the contents 
        as a nested dictionary where the highest level key is the file name. 

        Parameters
        ----------
        path : str
            Relative or absolute path to the FRA file.

        Returns
        -------
        fra_dict : dict
            The Maccor FRA file returned as a nested dictionary. 
        """
        parser = configparser.ConfigParser()
        parser.read(path)
        fra_dict = {section: dict(parser.items(section))
                    for section in parser.sections()}

        return {os.path.split(path)[-1]: fra_dict}

    def __fastwave_from_file(self, path) -> dict:
        """
        Reads the Maccor fastwave file (*.MFW) and returns the contents 
        as a nested dictionary where the highest level key is the file name
        and the corresponding value is the file contents as a string.

        Parameters
        ----------
        path : str
            Relative or absolute path to the fastwave file.

        Returns
        -------
        fra_dict : dict
            The Maccor fastwave file returned as a nested dictionary. 
        """
        with open(path, 'r') as file:
            fastwave_string = file.read()
        return {os.path.split(path)[-1]: fastwave_string}

    def __steps_from_procedure(self, procedure: dict) -> dict:
        """
        Determines which steps are charge/discharge/rest from the passed Maccor procedure
        and returns them as a dictionary of lists.

        Parameters
        ----------
        schedule : dict
            A nested dictionary containing the Maccor procedure. Assumes a layout generated
            by `__procedure_from_file()`

        Returns
        -------
        steps : dict
            A dict containing lists of the charge (key->'chg'), discharge, (key->'dsg'), and 
            rest (key->'rst') steps.
        """
        chg_steps = []
        dsg_steps = []
        rest_steps = []

        step_dict_list = procedure['ProcSteps']['TestStep']
        for i, step in enumerate(step_dict_list):
            if step['StepType'] in Constants.MACCOR_CHARGE_STEP_NAMES:
                chg_steps.append(i+1)
            elif step['StepType'] in Constants.MACCOR_DISCHARGE_STEP_NAMES:
                dsg_steps.append(i+1)
            elif step['StepType'] == 'Rest':
                rest_steps.append(i+1)

        steps = {'chg': chg_steps, 'dsg': dsg_steps, 'rst': rest_steps}
        logger.info("Identified Maccor procedure steps: " + json.dumps(steps))

        return steps

    def __arbin_schedule_from_files(self, paths) -> DashOrderedDict:
        """
        Reads the Arbin procedure and associated files the passed file path list.
        Returns all files as dictionaries in a single nested dictionary. Supported files to
        include in the paths list are Arbin schedules (*.sdx or *.sdu) (required),
        object files (*.to), simulation files (*.txt), CAN BMS files (*.can), and mapping 
        files (*.bth).

        Parameters
        ----------
        paths : list[str]
            Relative or absolute path to the schedule and associated files. 

        Returns
        -------
        procedure : DashOrderedDict
            The Maccor procedure and associated files saved as a nested dict.
        """
        schedule_dod = DashOrderedDict()
        object_file_dict = {}
        can_bms_file_dict = {}
        can_fm_file_dict = {}
        mapping_file_dict = {}
        simulation_dicts = []

        if self.cycler_make == '':
            self.cycler_make = Constants.MAKE_ARBIN
        elif self.cycler_make != Constants.MAKE_ARBIN:
            logger.error(
                "Passed schedule does not match previously assigned cycler_make!")
            return {}

        for path in paths:
            if path.endswith('.sdx') or path.endswith('.sdu'):
                if not self.schedule['file_name']:
                    self.schedule['file_name'] = path
                    logger.info("Importing Arbin schedule " +
                                self.schedule['file_name'])
                    schedule = self.__schedule_from_file(path)
                else:
                    logger.error("A schedule file is already defined!")
            elif path.endswith('.to'):
                if not object_file_dict:
                    logger.info("Importing Arbin object file " +
                                path)
                    object_file_dict = self.__object_from_file(path)
                else:
                    logger.error("An object file is already defined!")
            elif path.endswith('.can'):
                if not can_bms_file_dict:
                    logger.info("Importing Arbin CAN BMS file " +
                                path)
                    can_bms_file_dict = self.__can_bms_from_file(path)
                else:
                    logger.error("A CAN BMS file is already defined")
            elif path.endswith('.fm'):
                if not can_fm_file_dict:
                    logger.info("Importing Arbin CAN formula file " +
                                path)
                    can_fm_file_dict = self.__can_fm_from_file(path)
                else:
                    logger.error("A CAN formula file is already defined")
            elif path.endswith('.bth'):
                if not mapping_file_dict:
                    logger.info("Importing Arbin mapping file " +
                                path)
                    mapping_file_dict = self.__mapping_from_file(path)
                else:
                    logger.error("A mapping file is already defined")
            elif path.endswith('.txt'):
                logger.info("Importing Arbin simulation file " +
                            path)
                simulation_dicts.append(self.__simulation_from_file(path))
            else:
                logger.warning("Unknown file type included: " + str(path))

        try:
            schedule_dod.set('ArbinSchedule', schedule)
        except NameError:
            logger.error(
                "Missing Arbin schedule file from included paths!" + paths)
            return {}

        self.schedule['steps'] = self.__steps_from_schedule(schedule)

        if object_file_dict:
            schedule_dod.set('object_file', object_file_dict)
        if can_bms_file_dict:
            schedule_dod.set('can_bms_file', can_bms_file_dict)
        if can_fm_file_dict:
            schedule_dod.set('can_formula_file', can_fm_file_dict)
        if mapping_file_dict:
            schedule_dod.set('mapping_file', mapping_file_dict)
        if simulation_dicts:
            schedule_dod.set('simulation_files', simulation_dicts)

        return schedule_dod

    def __schedule_from_file(self, path) -> DashOrderedDict:
        """
        Reads the Arbin schedule (*.sdu or *.sdx) file and returns the contents 
        as a nested DashOrderedDict. This function was pulled from BEEP:
        https://github.com/TRI-AMDD/beep/blob/master/beep/protocol/arbin.py

        Parameters
        ----------
        path : str
            Relative or absolute path to the schedule file.

        Returns
        -------
        schedule : DashOrderedDict
            The Arbin schedule as a nested dictionary.
        """
        schedule = DashOrderedDict()

        with open(path, "rb") as f:
            text = f.read()
        text = text.decode(Constants.ARBIN_SCHEDULE_FILE_ENCODING)
        split_text = re.split(r"\[(.+)\]", text)

        for heading, body in zip(split_text[1::2], split_text[2::2]):
            body_lines = re.split(r"[\r\n]+", body.strip())
            body_dict = OrderedDict([line.split("=", 1)
                                    for line in body_lines])
            heading = heading.replace("_", ".")
            schedule.set(heading, body_dict)

        return schedule

    def __object_from_file(self, path) -> dict:
        """
        Reads the Arbin object file (*.to) and returns the contents 
        as a nested dictionary where top level key is the file name.

        Parameters
        ----------
        path : str
            Relative or absolute path to the object file.

        Returns
        -------
        object_file_dict : dict
            The Arbin object file returned as a dictionary. 
        """
        parser = configparser.ConfigParser()
        parser.read(path)
        confdict = {section: dict(parser.items(section))
                    for section in parser.sections()}
        return {os.path.split(path)[-1]: confdict}

    def __can_bms_from_file(self, path) -> OrderedDict:
        """
        Reads the Arbin CAN BMS file (*.to) and returns the contents 
        as a nested dictionary.

        Parameters
        ----------
        path : str
            Relative or absolute path to the CAN BMS file.

        Returns
        -------
        can_bms_dict: OrderedDict
            The CAN BMS file returned as a dictionary. 
        """
        with open(path, "rb") as f:
            text = f.read().decode('UTF-8')
            can_dict = xmltodict.parse(
                text, process_namespaces=False, strip_whitespace=True)
        return {os.path.split(path)[-1]: can_dict}

    def __can_fm_from_file(self, path) -> OrderedDict:
        """
        Reads the Arbin CAN formula file (*.fm) and returns the contents 
        as a nested dictionary where top level key is the file name.

        Parameters
        ----------
        path : str
            Relative or absolute path to the CAN formula file.

        Returns
        -------
        can_bms_dict: OrderedDict
            The CAN formula file returned as a dictionary. 
        """
        parser = configparser.ConfigParser()
        parser.read(path)
        fmdict = {section: dict(parser.items(section))
                  for section in parser.sections()}
        return {os.path.split(path)[-1]: fmdict}

    def __mapping_from_file(self, path) -> dict:
        """
        Reads the Arbin mapping file file (*.bth) and returns the contents 
        as a nested dictionary where top level key is the file name.

        Parameters
        ----------
        path : str
            Relative or absolute path to the mapping file.

        Returns
        -------
        mapping_file_dict: dict
            The mapping file returned as a dictionary. 
        """
        parser = configparser.ConfigParser()
        # Note: This is not the exact encoding and some special characters are being read incorrectly, but most of the file is being read correctly.
        parser.read(path, encoding='ISO-8859-1')
        mapping_dict = {section: dict(parser.items(section))
                        for section in parser.sections()}
        return {os.path.split(path)[-1]: mapping_dict}

    def __simulation_from_file(self, path) -> dict:
        """
        Reads the Arbin simulation file (*.MFW) and returns the contents 
        as a dictionary where the key is the file name and the corresponding
        value is the file contents as a string. 

        Parameters
        ----------
        path : str
            Relative or absolute path to the simulation file.

        Returns
        -------
        simulation_dict : dict
            The Arbin simulation file returned as a nested dictionary. 
        """
        with open(path, 'r') as file:
            simulation_string = file.read()
        return {os.path.split(path)[-1]: simulation_string}

    def __steps_from_schedule(self, schedule: DashOrderedDict) -> dict:
        """
        Determines which steps are charge/discharge/rest from the passed Arbin schedule
        and returns them as a dictionary of lists.

        Parameters
        ----------
        schedule : DashOrderedDict
            A nested DashOrderedDict containing the Arbin schedule. Assumes a layout generated
            by `__schedule_from_file()`

        Returns
        -------
        steps : dict
            A dict containing lists of the charge (key->'chg'), discharge, (key->'dsg'), and 
            rest (key->'rst') steps.
        """
        chg_steps = []
        dsg_steps = []
        rest_steps = []

        step_dicts = {
            key: val for key, val in schedule['Schedule'].items() if key.startswith('Step')}

        for i, step in enumerate(step_dicts):
            step_num = i+1
            step_type = step_dicts[step]['m_szStepCtrlType']
            if step_type == 'Current(A)':
                ctrl_value = step_dicts[step]['m_szCtrlValue']
                if ctrl_value[0] == '-':
                    logger.info("Categorizing 'Current(A)' control value " +
                                ctrl_value+" as discharge, step " + str(step_num))
                    dsg_steps.append(step_num)
                else:
                    logger.info("Categorizing 'Current(A)' control value " +
                                ctrl_value+" as charge, step " + str(step_num))
                    chg_steps.append(step_num)
            elif step_type == 'Current Ramp(A)':
                ctrl_value = step_dicts[step]['m_szCtrlValue']
                if ctrl_value[0] == '-':
                    logger.info("Categorizing 'Current Ramp(A)' control value " +
                                ctrl_value+" as discharge, step " + str(step_num))
                    dsg_steps.append(step_num)
                else:
                    logger.info("Categorizing 'Current Ramp(A)' control value " +
                                ctrl_value+" as charge, step " + str(step_num))
                    chg_steps.append(step_num)
            elif step_type == 'CCCV':
                logger.info(
                    "Categorizing 'CCCV' as charge, step " + str(step_num))
                chg_steps.append(step_num)
            elif step_type == 'CV':
                logger.info(
                    "Categorizing 'CV' as charge, step " + str(step_num))
                chg_steps.append(step_num)
            elif step_type == 'Current Simulation':
                logger.info("Categorizing 'Current Simulation' control value " +
                            ctrl_value+" as charge, step " + str(step_num))
                chg_steps.append(step_num)
            elif step_type == 'Rest':
                logger.info(
                    "Categorizing 'Rest' as rest, step " + str(step_num))
                rest_steps.append(step_num)

        steps = {'chg': chg_steps, 'dsg': dsg_steps, 'rst': rest_steps}
        logger.info("Identified Maccor procedure steps: " + json.dumps(steps))

        return steps

    def __data_from_file(self, path: str) -> pd.DataFrame:
        """
        Reads data from the passed file path and returns it as a pandas DataFrame.

        Parameters
        ----------
        path : str
            Relative or absolute path to the datafile. 

        Returns
        -------
        df : pandas.DataFrame
            A pandas DataFrame containing the data file.
        """
        logger.info(f'Load file path: {path}')
        if not os.path.exists(path):
            raise FileNotFoundError(f'Unable to load file {path}')

        headerLines, headerInfo = self.__get_header_lines(path)
        logger.debug(f'header lines: {headerLines}')
        logger.debug(f'header info: {headerInfo}')

        # Arbin Global Info
        if headerLines == -1:
            logger.debug('Arbin Global Info')
            self.raw_test_data_meta_data.append(headerInfo)
            logger.debug('Update raw_test_data_meta_data')
            self.cycler_make = Constants.MAKE_ARBIN
            return pd.DataFrame()

        df = pd.read_csv(
            path,
            skiprows=headerLines,
            sep='\t' if headerLines > 0 else ',',
            skipinitialspace=True,
            index_col=False,
            encoding_errors='replace'
        )

        logger.debug(
            f'Read {df.shape[0]} rows and {df.shape[1]} columns from {path}')

        df.columns = df.columns.str.strip()

        cycleMake, dataType = Utils.get_cycle_make(df.columns)
        logger.info(f'Cycle make: {cycleMake}. Data type: {dataType}')

        if cycleMake and dataType:
            self.cycler_make = cycleMake

            # meta data
            if cycleMake == Constants.MAKE_MACCOR:
                if dataType == Constants.DATA_TYPE_TEST_DATA:
                    self.raw_test_data_meta_data.append(headerInfo)
                    logger.debug('Update raw_test_data_meta_data')
                if dataType == Constants.DATA_TYPE_CYCLE_STATS:
                    self.raw_cycle_stats_meta_data.append(headerInfo)
                    logger.debug('Update raw_cycle_stats_meta_data')

            # test data
            if dataType == Constants.DATA_TYPE_TEST_DATA:
                self.raw_test_data = pd.concat(
                    [
                        self.raw_test_data,
                        df
                    ],
                    ignore_index=True
                )
                logger.debug(
                    f'Update raw_test_data. Total rows: {self.raw_test_data.shape[0]}')

            if dataType == Constants.DATA_TYPE_CYCLE_STATS:
                self.raw_cycle_stats = pd.concat(
                    [
                        self.raw_cycle_stats,
                        df
                    ],
                    ignore_index=True
                )
                logger.debug(
                    f'Update raw_cycle_stats. Total rows: {self.raw_cycle_stats.shape[0]}')

        return df

    def __get_header_lines(self, path: str) -> tuple[int, dict]:
        """
        Calculate header lines and extract info

        Parameters
        ----------
            path : str
                Test data file path

        Returns
        -------
            count : int
                Header lines
            header : int
                Header info
        """

        file = open(path, 'r', errors='replace')
        count = 0
        header = {}

        while True:
            line = file.readline()

            # The first line of the Arbin GlobalInfo file will be "TEST REPORT"
            if 'test report' in line.lower():
                file.close()
                return -1, self.__read_global_info(path)

            line = line.replace('\n', '').split('\t')
            line = list(filter(None, line))

            lineLen = len(line)
            if lineLen == 2 or lineLen == 0 or line[0][-1] == ':' or line in [
                ['Charge'], ['Discharge']
            ]:
                count += 1

                # Extract header info
                if lineLen == 2:
                    key = line[0][:-2] if line[0][-1] == ':' else line[0]
                    value = line[1]
                    header[key] = value
            else:
                break

        file.close()
        return count, header

    def __read_global_info(self, path):
        """
        Method for Arbin GlobalInfo file

        Parameters
        ----------
            path : str
                GlobalInfo data file path

        Returns
        -------
            header : int
                Header info
        """
        logger.debug('Read GlobalInfo')
        header = {}
        file = open(path, 'r')
        lines = []
        while True:
            line = file.readline()
            line = line.replace('\n', '').split(',')
            if line[0] == '':
                line = list(filter(None, line))
            lines.append(line)

            if not line:
                break

        file.close()

        # Hard-coding for the current situation
        # Test Name
        header[lines[1][0].strip()] = lines[1][1].strip()
        # Export Time
        header[lines[2][0].strip()] = lines[2][1].strip()
        # Serial Number
        header[lines[1][2].strip()] = lines[2][2].strip()
        # Other Info
        for i in range(len(lines[3])):
            header[lines[3][i].strip()] = lines[4][i].strip()

        return header

    def from_pickle(self, path) -> pd.DataFrame:
        """
        Reads data from the passed file path and returns it as a pandas DataFrame.

        Parameters
        ----------
        path : str
            Relative or absolute path to the datafile. 

        Returns
        -------
        df : pandas.DataFrame
            A pandas DataFrame containing the data file.
        """
        logger.info(f'Load pickle path: {path}')
        if not os.path.exists(path):
            raise FileNotFoundError(f'Unable to load file {path}')

        df = pd.read_pickle(path)
        logger.debug(
            f'Read {df.shape[0]} rows and {df.shape[1]} columns from {path}')

        return df



import re
import numpy as np
import pandas as pd
from typing import Callable

from battetl import logger, Constants, Utils


class Transformer:
    def __init__(
            self,
            timezone: str = None,
            user_transform_test_data: Callable[[
                pd.DataFrame], pd.DataFrame] = None,
            user_transform_cycle_stats: Callable[[
                pd.DataFrame], pd.DataFrame] = None) -> None:
        """
        An interface to transform battery test data to BattETL schema.

        Parameters
        ----------
        timezone : str
            Time zone strings in the IANA Time Zone Database. Used to convert to unix timestamp
            in seconds. Default 'America/Los_Angeles'.
        user_transform_test_data : Callable[[pd.DataFrame], pd.DataFrame], optional
            A user defined function to transform test data. The function should take a pandas.DataFrame
            as input and return a pandas.DataFrame as output.
        user_transform_cycle_stats : Callable[[pd.DataFrame], pd.DataFrame], optional
            A user defined function to transform cycle stats. The function should take a pandas.DataFrame
            as input and return a pandas.DataFrame as output.
        """
        # Default 'America/Los_Angeles'.
        self.timezone = timezone if timezone else Constants.DEFAULT_TIME_ZONE
        self.user_transform_test_data = user_transform_test_data
        self.user_transform_cycle_stats = user_transform_cycle_stats
        if self.user_transform_test_data:
            logger.info('User defined transform_test_data function found')
        if self.user_transform_cycle_stats:
            logger.info('User defined transform_cycle_stats function found')

        self.test_data = pd.DataFrame(dtype=object)
        self.cycle_stats = pd.DataFrame(dtype=object)

    def transform_test_data(self, data: pd.DataFrame, file_meta: dict = None) -> pd.DataFrame:
        """
        Transforms test data to conform to BattETL naming and data conventions

        Parameters
        ----------
        data : pandas.DataFrame
            The input DataFrame
        schedule_steps : dict
            A dictionary containing lists of charge (key->'chg'), discharge (key->'dsg'), and 
            rest (key->'rst') steps from the schedule used to generate the data. Used to calculate
            cycle level statistics (e.g. CV charge time.)s
        file_meta : dict, optional
            Dictionary containing the user defined column names for the test data. The default is None.
        Returns
        -------
        df : pandas.DataFrame
            The transformed output DataFrame
        """
        logger.info('Transform test data')

        df = data.copy()
        df = Utils.drop_unnamed_columns(df)
        df = Utils.drop_empty_rows(df)

        cycleMake, dataType = Utils.get_cycle_make(df.columns)
        logger.info(f'Cycle make: {cycleMake}. Data type: {dataType}')

        if file_meta:
            df = self.__transform_unstructured_data(df, file_meta)
        elif cycleMake == Constants.MAKE_ARBIN and dataType == Constants.DATA_TYPE_TEST_DATA:
            df = self.__transform_arbin_test_data(df)
        elif cycleMake == Constants.MAKE_MACCOR and dataType == Constants.DATA_TYPE_TEST_DATA:
            df = self.__transform_maccor_test_data(df)

        df = self.__consolidate_temps(df)

        # Apply user defined transformation
        if self.user_transform_test_data:
            df = self.user_transform_test_data(df)

        self.test_data = df.astype(object)
        return df

    def transform_cycle_stats(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Transforms cycle stats to conform to BattETL naming and data conventions

        Parameters
        ----------
        data : pandas.DataFrame
            The input DataFrame

        Returns
        -------
        df : pandas.DataFrame
            The transformed output DataFrame
        """
        logger.info('Transform cycle stats')

        df = data.copy()
        df = Utils.drop_unnamed_columns(df)
        df = Utils.drop_empty_rows(df)

        cycleMake, dataType = Utils.get_cycle_make(df.columns)
        logger.info(f'cycle make: {cycleMake}, data type: {dataType}')

        if cycleMake == Constants.MAKE_ARBIN and dataType == Constants.DATA_TYPE_CYCLE_STATS:
            df = self.__transform_arbin_cycle_stats(df)

        if cycleMake == Constants.MAKE_MACCOR and dataType == Constants.DATA_TYPE_CYCLE_STATS:
            df = self.__transform_maccor_cycle_stats(df)

        # Apply user defined transformation
        if self.user_transform_cycle_stats:
            df = self.user_transform_cycle_stats(df)

        self.cycle_stats = df.astype(object)
        return df

    def __transform_unstructured_data(self, df: pd.DataFrame, file_meta: dict) -> pd.DataFrame:
        """
        Transforms unstructured data with file_meta to conform to BattETL naming and data conventions
        1. Rename columns
        2. Concert units
        4. Convert data type
        5. Sort data
        Parameters
        ----------
        df : pandas.DataFrame
            The input DataFrame
        file_meta : dict
            Dictionary containing the meta data for the file.
        Returns
        -------
        df : pandas.DataFrame
            The transformed output DataFrame
        """
        logger.info('Transform unstructured data')

        # Check user defined column names exist
        if not file_meta.get('voltage_mv'):
            raise ValueError(
                f'Voltage column name does not exist: `voltage_mv`')
        if not file_meta.get('current_ma'):
            raise ValueError(
                f'Current column name does not exist: `current_ma`')
        if not file_meta.get('test_time_s') or file_meta.get('recorded_datetime'):
            raise ValueError(
                f'Required to have either `test_time_s` or `recorded_datetime` column!')

        # Remove the `pandas_read_csv_args` dictionary if it is there
        file_meta.pop("pandas_read_csv_args", None)

        # Rename columns
        columns_mappings = {}
        for column in file_meta:
            columns_mappings[file_meta[column]['column_name']] = column
        df = Utils.rename_df_columns(df, columnsMapping=columns_mappings)

        # Apply any transformations
        for column in file_meta:
            if file_meta.get(column).get('scaling_factor'):
                df[column] = df[column] * \
                    file_meta.get(column).get('scaling_factor')

        df = self.__convert_data_type(df)

        return df

    def __transform_arbin_test_data(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Transforms Arbin test data to conform to BattETL naming and data conventions
        1. Rename columns
        2. Convert to milli
        3. Convert datetime
        4. Convert data type
        5. Sort data

        Parameters
        ----------
        df : pandas.DataFrame
            The input DataFrame

        Returns
        -------
        df : pandas.DataFrame
            The transformed output DataFrame
        """
        df = Utils.rename_df_columns(
            df, Constants.COLUMNS_MAPPING_ARBIN_TEST_DATA)
        df = Utils.convert_to_milli(df)
        df = self.__convert_datetime_unixtime(df)
        df = self.__convert_data_type(df)
        df = Utils.sort_dataframe(df, ['unixtime_s', 'step'])

        return df

    def __transform_arbin_cycle_stats(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Transforms Arbin cycle stats to conform to BattETL naming and data conventions
        1. Rename columns
        2. Convert to milli
        3. Convert data type
        4. Sort data

        Parameters
        ----------
        df : pandas.DataFrame
            The input DataFrame

        Returns
        -------
        df : pandas.DataFrame
            The transformed output DataFrame
        """
        df = Utils.rename_df_columns(
            df, Constants.COLUMNS_MAPPING_ARBIN_CYCLE_STATS)
        df = Utils.convert_to_milli(df)
        df = self.__convert_data_type(df)
        df = Utils.sort_dataframe(df, ['cycle'])

        return df

    def __transform_maccor_test_data(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Transforms Maccor test data to conform to BattETL naming and data conventions
        1. Rename columns
        2. Convert to milli
        3. Convert datetime
        4. Convert data type
        5. Sort data

        Parameters
        ----------
        df : pandas.DataFrame
            The input DataFrame

        Returns
        -------
        df : pandas.DataFrame
            The transformed output DataFrame
        """
        df = Utils.rename_df_columns(
            df, Constants.COLUMNS_MAPPING_MACCOR_TEST_DATA)
        df = Utils.convert_to_milli(df)
        df = self.__convert_datetime_unixtime(df)
        df = self.__convert_data_type(df)

        if 'test_time_s' in df.columns and len(df) > 0 and self.__timedelta_validation_check(df['test_time_s'][0]):
            df = Utils.convert_timedelta_to_seconds(df, 'test_time_s')
        if 'step_time_s' in df.columns and len(df) > 0 and self.__timedelta_validation_check(df['step_time_s'][0]):
            df = Utils.convert_timedelta_to_seconds(df, 'step_time_s')

        df = Utils.sort_dataframe(df, ['unixtime_s', 'step'])

        return df

    def __transform_maccor_cycle_stats(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Transforms Maccor cycle to conform to BattETL naming and data conventions
        1. Rename columns
        2. Convert to milli
        3. Convert data type
        4. Convert test time
        5. Sort data

        Parameters
        ----------
        df : pandas.DataFrame
            The input DataFrame

        Returns
        -------
        df : pandas.DataFrame
            The transformed output DataFrame
        """
        df = Utils.rename_df_columns(
            df, Constants.COLUMNS_MAPPING_MACCOR_CYCLE_STATS)
        df = Utils.convert_to_milli(df)
        df = self.__convert_data_type(df)

        if 'test_time_s' in df.columns and len(df) > 0 and self.__timedelta_validation_check(df['test_time_s'][0]):
            df = Utils.convert_timedelta_to_seconds(df, 'test_time_s')

        df = Utils.sort_dataframe(df, ['cycle'])

        return df

    def __timedelta_validation_check(self, input_string):
        # Check if it is like the format "1d 15:07:52.77" or "1d 15:07:52"
        regex = re.compile(r'\d+d \d+:\d+:\d+(\.\d+)?\Z', re.I)
        match = regex.match(str(input_string))
        return bool(match)

    def __convert_datetime_unixtime(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Convert datetime to UTC format and add unixtime_s column

        Parameters
        ----------
        df : pandas.DataFrame
            The input DataFrame

        Returns
        -------
        df : pandas.DataFrame
            The transformed output DataFrame
        """
        logger.info('Convert datetime and add unixtime_s')

        df = Utils.convert_datetime(df, 'recorded_datetime', self.timezone)

        # Convert to unix timestamp
        df['unixtime_s'] = df['recorded_datetime'].astype(np.int64) // 10 ** 9

        return df

    def __convert_data_type(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Convert some data types if needed

        Parameters
        ----------
        df : pandas.DataFrame
            The input DataFrame

        Returns
        -------
        df : pandas.DataFrame
            The transformed output DataFrame
        """
        # Change the datatype to save on memory
        if 'cycle' in df.columns:
            logger.info('Convert column `cycle` to uint16')
            df['cycle'] = df['cycle'].astype('uint16')
        if 'stop' in df.columns:
            logger.info('Convert column `step` to uint8')
            df['step'] = df['step'].astype('uint8')

        # convert temperature columns to float
        temperature_columns = df.columns[
            df.columns.str.contains('thermocouple_')]
        if len(temperature_columns) > 0:
            logger.info('Convert temperature columns to float')
            for column in temperature_columns:
                df[column] = df[column].apply(Utils.convert_to_float)
        return df

    def __harmonize_capacity(self, df: pd.DataFrame, steps: dict) -> pd.DataFrame:
        """
        Harmonizes capacity/energy values across cyclers by creating four new columns: `charge_capacity_mah`, 
        `charge_energy_mwh`, `discharge_capacity_mah`, and `discharge_energy_mwh`. These columns only report 
        capacity/energy values during corresponding charge or discharge steps. Otherwise, they will be 
        `NaN`. The original capacity/energy reading is preserved in `cycler_charge_capacity_mah` or 
        `cycler_charge_energy_mwh` column, where `cycler` is replaced with the name of the cycler manufacturer.

        Parameters
        ----------
        df : pd.DataFrame
            A pandas DataFrame containing data for single battery cycle.
        steps : dict
            A dictionary containing lists of charge (key->'chg'), discharge (key->'dsg'), and rest (key->'rst') steps.

        Returns
        -------
        df : pd.DataFrame
            A pandas DataFrame with harmonized capacity/energy values across cyclers.
        """
        if 'arbin_charge_capacity_mah' in df:
            df['charge_capacity_mah'] = df[df.step.isin(
                steps['chg'])].arbin_charge_capacity_mah
            df['charge_energy_mwh'] = df[df.step.isin(
                steps['chg'])].arbin_charge_energy_mwh
            df['discharge_capacity_mah'] = df[df.step.isin(
                steps['dsg'])].arbin_discharge_capacity_mah
            df['discharge_energy_mwh'] = df[df.step.isin(
                steps['dsg'])].arbin_discharge_energy_mwh
            logger.debug(
                f'Harmonized capacity/energy values for Arbin cyclers. Modified {len(df)} rows and 4 columns.')
            logger.debug(
                'Added columns: charge_capacity_mah, charge_energy_mwh, discharge_capacity_mah, discharge_energy_mwh')
        elif 'maccor_capacity_mah' in df:
            df['charge_capacity_mah'] = df[df.step.isin(
                steps['chg'])].maccor_capacity_mah
            df['discharge_capacity_mah'] = df[df.step.isin(
                steps['dsg'])].maccor_capacity_mah
            if 'maccor_energy_mwh' in df:
                df['charge_energy_mwh'] = df[df.step.isin(
                    steps['chg'])].maccor_energy_mwh
                df['discharge_energy_mwh'] = df[df.step.isin(
                    steps['dsg'])].maccor_energy_mwh
                logger.debug(
                    f'Harmonized capacity/energy values for Maccor cyclers. Modified {len(df)} rows and 4 columns.')
                logger.debug(
                    'Added columns: charge_capacity_mah, charge_energy_mwh, discharge_capacity_mah, discharge_energy_mwh')
            else:
                logger.debug(
                    f'Harmonized capacity values for Maccor cyclers. Modified {len(df)} rows and 2 columns.')
                logger.debug(
                    'Added columns: charge_capacity_mah, discharge_capacity_mah')
        else:
            logger.warning("No capacity columns were found to refactor!")

        return df

    def calc_cycle_stats(self, steps: dict, cv_voltage_threshold_mv: float = None, cell_thermocouple: int = None) -> pd.DataFrame:
        """
        Calculates various charge and discharge statistics at the cycle level. Note this function 
        can only be run after self.test_data exists

        Parameters
        ----------
        steps : dict
            A dictionary containing lists of charge (key->'chg'), discharge (key->'dsg'), and 
            rest (key->'rst') steps.
        cv_voltage_thresh_mv : float
            The the voltage threshold in milli-volts above which charge is considered to be constant voltage.
        cell_thermocouple : int
            The number (as listed in the db column) of the thermocouple  that's attached to the cell.

        Returns
        -------
        self.cycle_stats : pd.DataFrame
            The calculated cycle statistics for all cycles.
        """
        if self.test_data.empty:
            logger.error("Cannot run `calc_cycle_stats()` without test_data!")
            return self.cycle_stats

        self.test_data = self.__harmonize_capacity(self.test_data, steps)

        # DataFrame where we will hold calculated cycle statistics for all cycles
        df_calced_stats = pd.DataFrame(columns=['cycle'])

        cycle_list = self.test_data['cycle'].unique()
        logger.info(
            f'Calculating cycle statistics for {len(cycle_list)} cycles')

        for cycle in cycle_list:
            cycle_data = self.test_data[self.test_data.cycle == cycle]
            if cycle_data.empty:
                logger.info("No cycle data for cycle " + str(cycle))
                continue

            # Calculate various charge and discharge cycle statistics.
            stats = {'cycle': cycle}

            charge_metrics = self.__calc_charge_stats(
                cycle_data, steps['chg'], cv_voltage_threshold_mv, cell_thermocouple)
            stats.update(charge_metrics)

            discharge_metrics = self.__calc_discharge_stats(
                cycle_data, steps['dsg'], cell_thermocouple)
            stats.update(discharge_metrics)

            # Calculate coulombic efficiency from the charge and discharge stats.
            if (('calculated_charge_capacity_mah' not in stats) or
                    ('calculated_discharge_capacity_mah' not in stats) or
                    (stats['calculated_charge_capacity_mah'] == 0)):
                ce = float("nan")
                logger.info(
                    "Unable to calculate coulombic efficiency for cycle " + str(cycle))
            else:
                ce = (discharge_metrics['calculated_discharge_capacity_mah']
                      / charge_metrics['calculated_charge_capacity_mah'])
            stats.update({'calculated_coulombic_efficiency': ce})

            # Append the cycle statistics from this cycle to our cumulative DataFrame for all cycles.
            df_calced_stats = pd.concat(
                [df_calced_stats, pd.DataFrame(stats, index=[0])], axis=0)

        if self.cycle_stats.empty:
            self.cycle_stats = df_calced_stats
        else:
            self.cycle_stats = self.cycle_stats.join(
                df_calced_stats.set_index('cycle'), on='cycle')

        return self.cycle_stats

    def __calc_charge_stats(self, cycle_data: pd.DataFrame, charge_steps: list, cv_voltage_threshold_mv: float = None, cell_thermocouple: int = None) -> dict:
        """
        Calculates various charge stats for a single cycle of data.

        Parameters
        ----------
        cycle_data : pd.DataFrame
            A DataFrame containing data for single battery cycle.
        charge_steps : list
            List of charge steps from the cycler schedule.
        cv_voltage_thresh_mv : float
            The the voltage threshold in milli-volts above which charge is considered to be constant voltage.
        cell_thermocouple : int
            The number (as listed in the db column) of the thermocouple  that's attached to the cell.

        Returns
        -------
        stats : dict
            A dictionary containing various charge stats.
        """
        logger.debug(
            f'Calculating charge statistics for cycle {cycle_data.cycle.iloc[0]} \
                with {len(cycle_data)} rows and {len(charge_steps)} charge steps.')

        if cv_voltage_threshold_mv:
            logger.debug(
                f'Using CV voltage threshold of {cv_voltage_threshold_mv} mV.')

        stats = {}

        # Define charge data to be where the step is a charge step.
        chg_data = cycle_data[cycle_data.step.isin(charge_steps)]
        if len(chg_data) < 2:
            logger.info("No charge data for cycle " +
                        str(cycle_data.cycle.iloc[0]))
            return stats

        ez_df = self.__ez_calc_df(
            chg_data, charge_steps, 'charge', cv_voltage_threshold_mv)

        stats['calculated_charge_capacity_mah'] = ez_df['charge_capacity_mah'].iloc[-1]
        stats['calculated_charge_energy_mwh'] = ez_df['charge_energy_mwh'].iloc[-1]
        stats['calculated_charge_time_s'] = ez_df['elapsed_time_s'].iloc[-1]

        stats['calculated_cc_charge_time_s'] = ez_df.cc_time_s.sum()
        stats['calculated_cv_charge_time_s'] = ez_df.cv_time_s.sum()
        stats['calculated_cc_capacity_mah'] = ez_df.cc_capacity_mah.sum()
        stats['calculated_cv_capacity_mah'] = ez_df.cv_capacity_mah.sum()

        # Calculate 50%/80% charge time & capacity.
        eighty_percent_cap_mah = stats['calculated_charge_capacity_mah'] * 0.8
        fifty_percent_cap_mah = stats['calculated_charge_capacity_mah'] * 0.5
        try:
            eighty_percent_cap_time_s = (ez_df[ez_df.charge_capacity_mah > eighty_percent_cap_mah].elapsed_time_s.iloc[0]
                                         - ez_df.elapsed_time_s.iloc[0])
            half_percent_cap_time_s = (ez_df[ez_df.charge_capacity_mah > fifty_percent_cap_mah].elapsed_time_s.iloc[0]
                                       - ez_df.elapsed_time_s.iloc[0])
        except:
            eighty_percent_cap_time_s = float('nan')
            half_percent_cap_time_s = float('nan')
            logger.warning(
                f'Incomplete charge data for cycle {cycle_data.cycle.iloc[0]}')

        stats['calculated_fifty_percent_charge_time_s'] = half_percent_cap_time_s
        stats['calculated_eighty_percent_charge_time_s'] = eighty_percent_cap_time_s

        if cell_thermocouple:
            logger.debug(
                f'Using cell thermocouple {cell_thermocouple} to calculate max charge temp.')
            thermocouple_col = f"thermocouple_{cell_thermocouple}_c"
            try:
                stats['calculated_max_charge_temp_c'] = chg_data.loc[:,
                                                                     thermocouple_col].max()
            except:
                logger.warning(
                    'cell_thermocouple value supplied, but not found in test_data.')

        return stats

    def __calc_discharge_stats(self, cycle_data: pd.DataFrame, discharge_steps: list, cell_thermocouple: int = None) -> dict:
        """
        Calculates various discharge stats for a single cycle of data.

        Parameters
        ----------
        cycle_data : pd.DataFrame
            A DataFrame containing data for single battery cycle.
        discharge_steps : list
            List of discharge steps from the cycler schedule.
        cell_thermocouple : int
            The number (as listed in the db column) of the thermocouple  that's attached to the cell.

        Returns
        -------
        stats : dict
            A dictionary containing various discharge stats.
        """
        logger.debug(
            f'Calculating discharge statistics for cycle {cycle_data.cycle.iloc[0]} \
                with {len(cycle_data)} rows and {len(discharge_steps)} discharge steps')
        stats = {}

        # Define discharge data to be where the step is a discharge step.
        dsg_data = cycle_data[cycle_data.step.isin(discharge_steps)]
        if len(dsg_data) < 2:
            logger.info("No discharge data for cycle " +
                        str(cycle_data.cycle.iloc[0]))
            return stats

        ez_df = self.__ez_calc_df(dsg_data, discharge_steps, 'discharge')

        stats['calculated_discharge_capacity_mah'] = ez_df['discharge_capacity_mah'].iloc[-1]
        stats['calculated_discharge_energy_mwh'] = ez_df['discharge_energy_mwh'].iloc[-1]
        stats['calculated_discharge_time_s'] = ez_df['elapsed_time_s'].iloc[-1]

        if cell_thermocouple:
            logger.debug(
                f'Using cell thermocouple {cell_thermocouple} to calculate max discharge temp.')
            thermocouple_col = f"thermocouple_{cell_thermocouple}_c"
            try:
                stats['calculated_max_discharge_temp_c'] = dsg_data.loc[:,
                                                                        thermocouple_col].max()
            except:
                logger.warning(
                    'cell_thermocouple value supplied, but not found in test_data.')
        return stats

    def __ez_calc_df(self, cycle_df: pd.DataFrame, steps: list, step_type: str, cv_voltage_thresh_mv: float = None) -> pd.DataFrame:
        """
        Creates a DataFrame we can easily use to calculate cycle statistics.

        Modifies test_data to cumulative capacity in cases where capacity is reset at each step.

        Parameters
        ----------
        df : pd.DataFrame
            The DataFrame use to calculate cumulative capacity.
        steps : list
            A list of the steps to calculate cumulative capacity for. 
        step_type : str
            Either 'charge' or 'discharge' depending on what type of steps we are calculating for.
        cv_voltage_thresh_mv : float
            The voltage threshold in mV above which charge is considered to be constant voltage.    

        Returns
        -------
        ez_df: pd.DataFrame
            DataFrame containing values filtered to charge steps with cumulative capacity
            and elapsed time calculated.
        """
        logger.debug(
            f'Calculating cumulative capacity with {len(cycle_df)} rows, \
                  {len(steps)} {step_type} steps and cv_voltage_thresh_mv: {cv_voltage_thresh_mv}')

        time_col = 'elapsed_time_s'
        volt_col = 'voltage_mv'
        cc_time = 'cc_time_s'
        cv_time = 'cv_time_s'
        cc_cap = 'cc_capacity_mah'
        cv_cap = 'cv_capacity_mah'
        if step_type == 'charge':
            cap_col = 'charge_capacity_mah'
            eng_col = 'charge_energy_mwh'
        elif step_type == 'discharge':
            cap_col = 'discharge_capacity_mah'
            eng_col = 'discharge_energy_mwh'
        else:
            logger.error(f'Unknown step type {step_type}!')
            return pd.DataFrame()

        ez_df = pd.DataFrame(
            columns=[time_col, volt_col, cap_col, eng_col, cc_time, cv_time, cc_cap, cv_cap])

        # filter step slice to charge/discharge steps
        relevant_steps = [step for step in cycle_df.step.unique()
                          if step in steps]

        # Iterate through each charge step to calculate cumulative capacity
        for step in relevant_steps:
            step_slice = cycle_df[cycle_df.step == step]

            if step_slice.empty:
                continue
            step_df = pd.DataFrame(
                columns=[time_col, volt_col, cap_col, eng_col, cc_time, cv_time, cc_cap, cv_cap])

            # Checks if we've processed at least one step already
            if not ez_df.empty:
                # Keeps running time across steps. Ignoring time between non-charge/discharge steps
                step_df[time_col] = \
                    (step_slice['test_time_s'] - step_slice['test_time_s'].iloc[0]) + ez_df[time_col].iloc[-1]
                
                # This catches if capacity was reset after each step. Modifies test_data DF in place.
                if step_slice[cap_col].iloc[0] < ez_df[cap_col].iloc[-1]:
                    current_cycle = cycle_df.cycle.iloc[0]
                    self.test_data.loc[
                        (self.test_data.cycle == current_cycle) &
                        (self.test_data.step == step),
                        cap_col] += ez_df[cap_col]
                    step_slice[cap_col] += ez_df[cap_col].iloc[-1]

                    if eng_col in self.test_data.columns:
                        self.test_data.loc[
                            (self.test_data.cycle == current_cycle) &
                            (self.test_data.step == step),
                            eng_col] += ez_df[eng_col]
                        step_slice[eng_col] += ez_df[eng_col].iloc[-1]
            else:
                step_df[time_col] = step_slice['test_time_s'] - \
                    step_slice['test_time_s'].iloc[0]

            # Step_slice is updated with above updates to test_data because it's a slice, not copy, of test_data
            step_df[volt_col] = step_slice[volt_col]
            step_df[cap_col] = step_slice[cap_col]
            if eng_col in step_slice.columns:
                step_df[eng_col] = step_slice[eng_col]

            if step_type == 'charge' and cv_voltage_thresh_mv is not None:
                latest_capacity = ez_df[cap_col].iloc[-1] if not ez_df.empty else 0

                # Calculate the delta time/cap for each step, sum the deltas in calc_stats() to get cycle charge time/cap
                delta_time = step_slice['step_time_s'] - \
                    step_slice['step_time_s'].shift(1, fill_value=0)
                step_df[cc_time] = np.where(
                    step_slice[volt_col] < cv_voltage_thresh_mv, delta_time, 0)
                step_df[cv_time] = np.where(
                    step_slice[volt_col] >= cv_voltage_thresh_mv, delta_time, 0)
                delta_cap = step_slice[cap_col] - step_slice[cap_col].shift(
                    1, fill_value=latest_capacity)
                step_df[cc_cap] = np.where(
                    step_slice[volt_col] < cv_voltage_thresh_mv, delta_cap, 0)
                step_df[cv_cap] = np.where(
                    step_slice[volt_col] >= cv_voltage_thresh_mv, delta_cap, 0)
            elif step_type == 'discharge':
                step_df[[cc_time, cc_cap, cv_time, cv_cap]] = np.nan
            ez_df = pd.concat([ez_df, step_df])

        return ez_df

    def __consolidate_temps(self, df):
        '''
        Adds thermocouple readings from each data point to an array containing all thermocouple readings.

        Parameters
        ----------
        df : pandas.DataFrame
            DataFrame containing test data.

        Returns
        -------
        pd.DataFrame
            DataFrame containing test data with thermocouple readings in an array.
        '''
        thermocouple_cols = [
            col for col in df.columns if re.search('thermocouple_\d+_c', col)]

        df['thermocouple_temps_c'] = df.apply(
            lambda row: [row[col] for col in thermocouple_cols], axis=1)

        return df



import os
import json
import pandas as pd
from typing import Callable

from battetl import logger, Utils
from battetl.extract import Extractor
from battetl.transform import Transformer
from battetl.load import Loader


class BattETL:
    def __init__(
            self, config_path: str,
            user_transform_test_data: Callable[[
                pd.DataFrame], pd.DataFrame] = None,
            user_transform_cycle_stats: Callable[[
                pd.DataFrame], pd.DataFrame] = None,
            env_path: str = os.path.join(os.getcwd(), '.env')):
        """
        Creates the a class instance. The `start()` method still needs to be run to create the connection
        and to check the validity of the config.

        Parameters
        ----------
        config_path : str
            Path to the json configuration file for the class instance. The configuration file must contain
            the following:
            - `test_name` - Test name should be define in the `test_data_meta` table.
            - `data_file_path` - Absolute or relative path to the test data file.
            - `stats_file_path` - Absolute or relative path to the cycle stats data file.
            - `schedule_file_path` - Absolute or relative path to the schedule or procedure file.
            - `meta_data` - Dictionary containing the meta data for the test.

        user_transform_test_data : Callable[[pd.DataFrame], pd.DataFrame], optional
            A user defined function to transform test data. The function should take a pandas.DataFrame
            as input and return a pandas.DataFrame as output.
        user_transform_cycle_stats : Callable[[pd.DataFrame], pd.DataFrame], optional
            A user defined function to transform cycle stats. The function should take a pandas.DataFrame
            as input and return a pandas.DataFrame as output.
        env_path : str, optional
            Path to the .env file containing the environment variables. The default is the current working
            directory.
        """
        self.env_path = env_path
        Utils.load_env(env_path)

        self.config = Utils.load_config(config_path)

        self.user_transform_test_data = user_transform_test_data
        self.user_transform_cycle_stats = user_transform_cycle_stats

        self.raw_test_data = pd.DataFrame()
        self.test_data = pd.DataFrame()
        self.raw_cycle_stats = pd.DataFrame()
        self.cycle_stats = pd.DataFrame()
        self.schedule = None

    def extract(self):
        """
        Extracts the data in the target directory specified in the config.

        Returns
        -------
        self : BattETL
            Returns a reference to the instance object
        """
        extractor = Extractor()

        # Test data
        if self.config.get('data_file_path'):
            try:
                extractor.data_from_files(self.config['data_file_path'])
                self.raw_test_data = extractor.raw_test_data
            except Exception as e:
                logger.error('Failed to extract test data', exc_info=True)
                logger.error(e)
        else:
            logger.warning('No test data file path')

        # Cycle stats
        if self.config.get('stats_file_path'):
            try:
                extractor.data_from_files(self.config['stats_file_path'])
                self.raw_cycle_stats = extractor.raw_cycle_stats
            except Exception as e:
                logger.error('Failed to extract cycle stats', exc_info=True)
                logger.error(e)
        else:
            logger.warning('No cycle stats file path')

        # Schedule
        if self.config.get('schedule_file_path'):
            try:
                extractor.schedule_from_files(
                    self.config['schedule_file_path'])
                self.schedule = extractor.schedule
                self.config['meta_data']['schedule_meta']['details'] = json.dumps(
                    extractor.schedule)
            except Exception as e:
                logger.error('Failed to extract schedule', exc_info=True)
                logger.error(e)
        else:
            logger.warning('No schedule file path')

        logger.info('Finished extracting data')

        return self

    def transform(self):
        """
        Transforms the test data from the target directory.

        Returns
        -------
        self : BattETL
            Returns a reference to the instance object
        """
        transformer = Transformer(
            timezone=self.config.get('timezone'),
            user_transform_test_data=self.user_transform_test_data,
            user_transform_cycle_stats=self.user_transform_cycle_stats)

        if not self.raw_test_data.empty:
            try:
                transformer.transform_test_data(self.raw_test_data)
                self.test_data = transformer.test_data
            except Exception as e:
                logger.error('Failed to transform test data', exc_info=True)
                logger.error(e)
        else:
            logger.warning('No test data to transform.')

        if not self.raw_cycle_stats.empty:
            try:
                transformer.transform_cycle_stats(self.raw_cycle_stats)
                self.cycle_stats = transformer.cycle_stats
            except Exception as e:
                logger.error('Failed to transform cycle stats', exc_info=True)
                logger.error(e)
        else:
            logger.warning('No cycle stats to transform.')

        if not self.test_data.empty and self.schedule:
            try:
                cv_voltage_threshold_mv = self.config['meta_data']['schedule_meta'].get(
                    'cv_voltage_threshold_mv')
                cell_thermocouple = self.config.get('cell_thermocouple')

                transformer.calc_cycle_stats(
                    self.schedule['steps'],
                    cv_voltage_threshold_mv=cv_voltage_threshold_mv,
                    cell_thermocouple=cell_thermocouple)
                self.cycle_stats = transformer.cycle_stats
            except Exception as e:
                logger.error('Failed to calculate cycle stats', exc_info=True)
                logger.error(e)
        else:
            logger.warning('Skipping cycle stats calculation.')

        logger.info('Finished transforming data')

        return self

    def load(self):
        """
        Loads the test data, cycle stats, and schedule file to the target databases(s) 
        specified in the config.

        Returns
        -------
        num_rows_inserted : int
            The number of rows inserted into the target_table. 
        """

        loader = Loader(
            config=self.config,
            env_path=self.env_path)

        # Load test data
        if not self.test_data.empty:
            num_rows_inserted_test_data = loader.load_test_data(self.test_data)
            logger.info(
                f'Loaded {num_rows_inserted_test_data} rows of test data to database!')
        else:
            logger.warning('No test data to load.')

        # Load cycle stats
        if not self.cycle_stats.empty:
            num_rows_inserted_cycle_stats = loader.load_cycle_stats(
                self.cycle_stats)
            logger.info(
                f'Loaded {num_rows_inserted_cycle_stats} rows of cycle stats to database')
        else:
            logger.warning('No cycle stats to load.')

        del loader

        logger.info('Finished loading data')

        return self



class Constants:
    DEFAULT_TIME_ZONE = 'America/Los_Angeles'

    BATTDB_SCHEMA_VERSION = 9.7
    BATTDB_QUICK_SCHEMA_VERSION = 1.1

    DATABASE_MAX_RETRIES = 10
    DATABASE_RETRY_DELAY = 10
    DATABASE_MAX_RETRY_DELAY = 60

    MAKE_ARBIN = 'arbin'
    MAKE_MACCOR = 'maccor'
    DATA_TYPE_TEST_DATA = 'test_data'
    DATA_TYPE_CYCLE_STATS = 'cycle_stats'

    MACCOR_PROCEDURE_FILE_ENCODING = 'UTF-8'

    MACCOR_CHARGE_STEP_NAMES = ['Charge', 'Chg Func', 'FastWave']
    # Yes, discharge is actually misspelled in the extract procedure files
    MACCOR_DISCHARGE_STEP_NAMES = ['Dischrge', 'Dis Func']

    ARBIN_SCHEDULE_FILE_ENCODING = 'latin-1'

    PREFIX_ARBIN_THERMOCOUPLE = 'aux_temperature_'
    PREFIX_MACCOR_THERMOCOUPLE = 'temp '
    TEMPLATE_RENAMED_THERMOCOUPLE = 'thermocouple_X_c'

    COLUMNS_TEST_DATA = {
        'test_data_id',
        'test_id',
        'cycle',
        'step',
        'test_time_s',
        'step_time_s',
        'current_ma',
        'voltage_mv',
        'recorded_datetime',
        'unixtime_s',
        'thermocouple_temps_c',
        'other_details',
    }
    COLUMNS_CYCLE_STATS = {
        'cycle_stats_id',
        'test_id',
        'cycle',
        'test_time_s',
        'reported_charge_capacity_mah',
        'reported_discharge_capacity_mah',
        'reported_charge_energy_mwh',
        'reported_discharge_energy_mwh',
        'reported_charge_time_s',
        'reported_discharge_time_s',
        'calculated_charge_capacity_mah',
        'calculated_max_charge_temp_c',
        'calculated_discharge_capacity_mah',
        'calculated_max_discharge_temp_c',
        'calculated_cc_charge_time_s',
        'calculated_cv_charge_time_s',
        'calculated_cc_capacity_mah',
        'calculated_cv_capacity_mah',
        'reported_coulombic_efficiency',
        'calculated_coulombic_efficiency',
        'calculated_fifty_percent_charge_time_s',
        'calculated_eighty_percent_charge_time_s',
        'calculated_charge_energy_mwh',
        'calculated_discharge_energy_mwh',
        'other_details',
    }
    COLUMNS_ARBIN_TEST_DATA_ONLY = {
        'Date Time',
        'ACR (Ohm)',
        'dQ/dV (Ah/V)',
        'Internal Resistance (Ohm)',
        'dV/dQ (V/Ah)',
        'dV/dt (V/s)',
        'Data Point',
    }
    COLUMNS_ARBIN_CYCLE_STATS_ONLY = {
        'Charge Time (s)',
        'Date_Time',
        'mAh/g',
        'Coulombic Efficiency (%)',
        'V_Max_On_Cycle (V)',
        'Discharge Time (s)',
    }
    COLUMNS_MACCOR_TEST_DATA_ONLY = {
        'Cyc#',
        'StepTime(s)',
        'DPt Time',
        'Current(A)',
        'Capacity(Ah)',
        'Step',
        'EV Temp',
        'Voltage(V)',
        'TestTime(s)',
        'Temp 1',
    }

    COLUMNS_MACCOR_TEST_DATA_TYPE2_ONLY = {
        'Rec',
        'Cycle P',
        'Cycle C',
        'Capacity',
        'Energy',
        'MD',
        'ES',
        'DPT Time',
    }

    COLUMNS_MACCOR_CYCLE_STATS_ONLY = {
        'T1_End',
        'T1_Max',
        'T1_Start',
        'T1_Min',
        'Cycle',
        'Date',
        'AH-OUT',
        'AH-IN',
    }

    COLUMNS_MACCOR_CYCLE_STATS_CUSTOMER1 = {
        'Cycle',
        'AH-IN',
        'AH-OUT',
        'T1_Start',
        'T1_End',
        'T1_Min',
        'T1_Max',
        'T1_Start',
        'T1_End',
        'T1_Min',
        'T1_Max',
        'Date',
    }

    COLUMNS_MACCOR_TEST_DATA_CUSTOMER1 = {
        'Cyc#',
        'Step',
        'TestTime(s)',
        'StepTime(s)',
        'Capacity(Ah)',
        'Watt-hr',
        'Current(A)'
        'Voltage(V)',
        'ES',
        'DPt Time',
        'Volt 1',
        'ManufacturerAccess (0x00)',
        'AtRate (0x02)',
        'AtRateTimeToEmpty (0x04)',
        'Temperature (0x06)',
        'Voltage (0x08)',
        'BatteryStatus (0x0A)',
        'Current (0x0C)',
        'RemainingCapacity (0x10)',
        'FullChargeCapacity (0x12)',
        'AverageCurrent (0x14)',
        'AverageTimeToEmpty (0x16)',
        'AverageTimeToFull (0x18)',
        'RelativeStateOfCharge (0x2C)',
        'ChargingVoltage (0x30)',
        'ChargingCurrent (0x32)',
        'DesignCapacity (0x3C)',
    }

    COLUMNS_TO_MILLI = {
        # Test Data
        'voltage_v': 'voltage_mv',
        'current_a': 'current_ma',
        'voltage': 'voltage_mv',
        'current': 'current_ma',
        'charge_capacity_ah': 'charge_capacity_mah',
        'discharge_capacity_ah': 'discharge_capacity_mah',
        'charge_energy_wh': 'charge_energy_mwh',
        'discharge_energy_wh': 'discharge_energy_mwh',
        'power_w': 'power_mw',
        'impedance_ohm': 'impedance_mohm',
        'capacity_ah': 'capacity_mah',
        # Arbin Test Data
        'arbin_charge_capacity_ah': 'arbin_charge_capacity_mah',
        'arbin_discharge_capacity_ah': 'arbin_discharge_capacity_mah',
        'arbin_charge_energy_wh': 'arbin_charge_energy_mwh',
        'arbin_discharge_energy_wh': 'arbin_discharge_energy_mwh',
        # Maccor Test Data
        'maccor_capacity_ah': 'maccor_capacity_mah',
        'maccor_energy_wh': 'maccor_energy_mwh',
        'capacity': 'maccor_capacity_mah',
        'energy': 'maccor_energy_mwh',
        # Arbin Cycle Stats
        'reported_charge_capacity_ah': 'reported_charge_capacity_mah',
        'reported_discharge_capacity_ah': 'reported_discharge_capacity_mah',
        'reported_charge_energy_wh': 'reported_charge_energy_mwh',
        'reported_discharge_energy_wh': 'reported_discharge_energy_mwh',
    }

    COLUMNS_MAPPING_ARBIN_TEST_DATA = {
        'Date_Time': 'recorded_datetime',
        'Internal Resistance (Ohm)': 'impedance_ohm',
        'Date_Time': 'recorded_datetime',
        'Date Time': 'recorded_datetime',
        'Data Point': 'data_point',
        'Test Time (s)': 'test_time_s',
        'Test Time(s)': 'test_time_s',
        'Test_Time(s)': 'test_time_s',
        'Step Time (s)': 'step_time_s',
        'Step_Time(s)': 'step_time_s',
        'Step Time(s)': 'step_time_s',
        'Cycle Index': 'cycle',
        'Cycle_Index': 'cycle',
        'Step Index': 'step',
        'Step_Index': 'step',
        'TC_Counter1': 'tc_counter1',
        'Voltage (V)': 'voltage_v',
        'Voltage(V)': 'voltage_v',
        'Current (A)': 'current_a',
        'Current(A)': 'current_a',
        'Charge Capacity (Ah)': 'arbin_charge_capacity_ah',
        'Charge Capacity(Ah)': 'arbin_charge_capacity_ah',
        'Charge_Capacity(Ah)': 'arbin_charge_capacity_ah',
        'Discharge Capacity (Ah)': 'arbin_discharge_capacity_ah',
        'Discharge Capacity(Ah)': 'arbin_discharge_capacity_ah',
        'Discharge_Capacity(Ah)': 'arbin_discharge_capacity_ah',
        'Charge Energy (Wh)': 'arbin_charge_energy_wh',
        'Charge_Energy (Wh)': 'arbin_charge_energy_wh',
        'Charge_Energy(Wh)': 'arbin_charge_energy_wh',
        'Discharge Energy (Wh)': 'arbin_discharge_energy_wh',
        'Discharge Energy(Wh)': 'arbin_discharge_energy_wh',
        'Discharge_Energy(Wh)': 'arbin_discharge_energy_wh',
        'Power (W)': 'power_w',
    }
    COLUMNS_MAPPING_ARBIN_CYCLE_STATS = {
        'Date_Time': 'recorded_datetime',
        'Test Time (s)': 'test_time_s',
        'Test Time(s)': 'test_time_s',
        'Step Time (s)': 'step_time_s',
        'Step Time(s)': 'step_time_s',
        'Cycle Index': 'cycle',
        'Step Index': 'step',
        'TC_Counter1': 'tc_counter1',
        'Voltage(V)': 'voltage_v',
        'Voltage (V)': 'voltage_v',
        'Current (A)': 'current_a',
        'Current(A)': 'current_a',
        'Charge Capacity (Ah)': 'reported_charge_capacity_ah',
        'Charge Capacity(Ah)': 'reported_charge_capacity_ah',
        'Discharge Capacity (Ah)': 'reported_discharge_capacity_ah',
        'Discharge Capacity(Ah)': 'reported_discharge_capacity_ah',
        'Charge Time (s)': 'reported_charge_time_s',
        'Charge Time(s)': 'reported_charge_time_s',
        'Discharge Time (s)': 'reported_discharge_time_s',
        'Coulombic Efficiency (%)': 'reported_coulombic_efficiency',
        'Charge Energy (Wh)': 'reported_charge_energy_wh',
        'Charge_Energy(Wh)': 'reported_charge_energy_wh',
        'Discharge Energy (Wh)': 'reported_discharge_energy_wh',
        'Discharge Energy(Wh)': 'reported_discharge_energy_wh',
    }
    COLUMNS_MAPPING_MACCOR_TEST_DATA = {
        'Cyc#': 'cycle',
        'Cycle P': 'cycle',
        'Step': 'step',
        'TestTime(s)': 'test_time_s',
        'Test Time': 'test_time_s',
        'StepTime(s)': 'step_time_s',
        'Step Time': 'step_time_s',
        'Capacity(Ah)': 'maccor_capacity_ah',
        'Watt-hr': 'maccor_energy_wh',
        'Current(A)': 'current_a',
        'Voltage(V)': 'voltage_v',
        'DPt Time': 'recorded_datetime',
        'EV Temp': 'ev_temp_c',
    }
    COLUMNS_MAPPING_MACCOR_CYCLE_STATS = {
        'Cycle': 'cycle',
        'Test Time': 'test_time_s',
        'Current': 'maccor_min_current_ma',
        'Voltage': 'maccor_min_voltage_mv',
        'AH-IN': 'reported_charge_capacity_ah',
        'AH-OUT': 'reported_discharge_capacity_ah',
        'WH-IN': 'reported_charge_energy_wh',
        'WH-OUT': 'reported_discharge_energy_wh',
        'T1_Start': 'maccor_charge_thermocouple_start_c',
        'T1_End': 'maccor_charge_thermocouple_end_c',
        'T1_Min': 'maccor_charge_thermocouple_min_c',
        'T1_Max': 'maccor_charge_thermocouple_max_c',
        'T1_Start.1': 'maccor_discharge_thermocouple_start_c',
        'T1_End.1': 'maccor_discharge_thermocouple_end_c',
        'T1_Min.1': 'maccor_discharge_thermocouple_min_c',
        'T1_Max.1': 'maccor_discharge_thermocouple_max_c',
        'ACR': 'acr_ohm',
    }
 
    COLUMNS_UNSTRUCTURED_TEST_DATA = {
        'voltage_mv',
        'current_ma',
        'time_s',
        'other_details',
    }
    UNSTRUCTURED_DATA_REQUIRED_KEYS = {
        'voltage_mv',
        'current_ma',
        'pandas_read_csv_args'
    }



import os
import json
import pytest
import pandas as pd
from copy import deepcopy

from battetl.load import Loader, BattDbTestHelper

CONFIG_DIR = os.path.join(os.path.dirname(__file__), 'configs')


class Values:
    """
    Class to store values that are used in multiple tests

    Attributes
    ----------
    CONFIG_1: dict
        The configuration dictionary for the test
    TEST_HELPER: BattDbTestHelper
        The BattDbTestHelper object that is used in the tests
    """
    CONFIG_1: dict = None
    TEST_HELPER: BattDbTestHelper = None


@pytest.fixture(scope='module', autouse=True)
def add_database_test_entries():
    """
    Fixture to add test entries to the database before the tests are run
    and delete them after the tests are run
    """
    with open(os.path.join(CONFIG_DIR, 'config_1.json')) as config_file:
        Values.CONFIG_1 = json.load(config_file)

    Values.TEST_HELPER = BattDbTestHelper(Values.CONFIG_1)

    # Will be executed before the first test
    Values.TEST_HELPER.create_test_db_entries()
    yield Values.TEST_HELPER
    # Will be executed after the last test
    Values.TEST_HELPER.delete_test_db_entries()


@pytest.mark.database
@pytest.mark.load
def test_lookup_cell_type_id():

    config = deepcopy(Values.CONFIG_1)

    # Lookup existing cell_type_id
    loader = Loader(config)
    cell_type_id = loader._Loader__lookup_cell_type_id()
    assert (cell_type_id == Values.TEST_HELPER.cell_type_id)
    del loader

    # Lookup non-existent cell_type_id
    config['meta_data']['cell_meta']['manufacturer'] = Values.TEST_HELPER.generate_random_string()
    config['meta_data']['cell_meta']['manufacturer_pn'] = Values.TEST_HELPER.generate_random_string()
    loader = Loader(config)
    cell_type_id = loader._Loader__lookup_cell_type_id()
    assert (cell_type_id is None)


@pytest.mark.database
@pytest.mark.load
def test_lookup_cell_id():

    config = deepcopy(Values.CONFIG_1)

    # Lookup existing cell id
    loader = Loader(config)
    cell_id = loader._lookup_cell_id()
    assert (cell_id == Values.TEST_HELPER.cell_id)
    del loader

    # Lookup non-existent cell_id
    config['meta_data']['cell']['manufacturer_sn'] = Values.TEST_HELPER.generate_random_string()
    config['meta_data']['cell']['label'] = Values.TEST_HELPER.generate_random_string()
    loader = Loader(config)
    cell_id = loader._lookup_cell_id()
    assert (cell_id is None)


@pytest.mark.database
@pytest.mark.load
def test_lookup_schedule_id():

    config = deepcopy(Values.CONFIG_1)

    # Lookup existing schedule id
    loader = Loader(config)
    schedule_id = loader._Loader__lookup_schedule_id()
    assert (schedule_id == Values.TEST_HELPER.schedule_id)
    del loader

    # Lookup non-existent schedule_id
    config['meta_data']['schedule_meta']['schedule_name'] = Values.TEST_HELPER.generate_random_string()
    loader = Loader(config)
    schedule_id = loader._Loader__lookup_schedule_id()
    assert (schedule_id is None)


@pytest.mark.database
@pytest.mark.load
def test_lookup_cycler_type_id():

    config = deepcopy(Values.CONFIG_1)

    # Lookup existing cycler type id
    loader = Loader(config)
    cycler_type_id = loader._Loader__lookup_cycler_type_id()
    assert (cycler_type_id == Values.TEST_HELPER.cycler_type_id)
    del loader

    # Lookup non-existent schedule_id
    config['meta_data']['cycler_meta']['manufacturer'] = Values.TEST_HELPER.generate_random_string()
    config['meta_data']['cycler_meta']['model'] = Values.TEST_HELPER.generate_random_string()
    loader = Loader(config)
    cycler_type_id = loader._Loader__lookup_cycler_type_id()
    assert (cycler_type_id is None)
    del loader


@pytest.mark.database
@pytest.mark.load
def test_lookup_cycler_id():

    config = deepcopy(Values.CONFIG_1)

    # Lookup existing cycler id
    loader = Loader(config)
    cycler_id = loader._lookup_cycler_id()
    assert (cycler_id == Values.TEST_HELPER.cycler_id)
    del loader

    # Lookup non-existing cycler with existing cycler_type_id
    config['meta_data']['cycler']['sn'] = Values.TEST_HELPER.generate_random_string()
    loader = Loader(config)
    cycler_id = loader._lookup_cycler_id()
    assert (cycler_id is None)
    del loader


@pytest.mark.database
@pytest.mark.load
def test_lookup_test_id():

    config = deepcopy(Values.CONFIG_1)

    # Lookup existing test id
    loader = Loader(config)
    test_id = loader._lookup_test_id()
    assert (test_id == Values.TEST_HELPER.test_id)
    del loader

    # Lookup non-existent test.
    config['meta_data']['test_meta']['test_name'] = Values.TEST_HELPER.generate_random_string()
    loader = Loader(config)
    test_id = loader._lookup_test_id()
    assert (test_id is None)
    del loader


@pytest.mark.database
@pytest.mark.load
def test_lookup_latest_unixtime():

    config = deepcopy(Values.CONFIG_1)

    Values.TEST_HELPER.delete_test_data()

    # Lookup unixtime_s for an existing test with no data in data table.
    loader = Loader(config)
    latest_unixtime_s = loader._Loader__lookup_latest_unixtime()
    assert (latest_unixtime_s is None)

    # Define data to upload
    data_dict = {
        'test_id': [Values.TEST_HELPER.test_id, Values.TEST_HELPER.test_id],
        'cycle': [1, 1],
        'step': [1, 1],
        'test_time_s': [1.0, 2.0],
        'step_time_s': [1.0, 2.0],
        'current_ma': [0.0, 0.0],
        'voltage_mv': [3.789, 3.800],
        'recorded_datetime': [pd.Timestamp(1674659265, unit='s', tz='US/Pacific'),
                              pd.Timestamp(1674659266, unit='s', tz='US/Pacific')],
        'unixtime_s': [1674659265.0, 1674659266.0]
    }
    df = pd.DataFrame(data=data_dict)

    # Load data to the table.
    num_rows_inserted = Values.TEST_HELPER.load_df_to_db(
        df=df, target_table='test_data')
    assert (num_rows_inserted == 2)

    # Check latest unixtims_s matches loaded data.
    latest_unixtime_s = loader._Loader__lookup_latest_unixtime()
    assert (latest_unixtime_s == data_dict['unixtime_s'][-1])

    del loader
    Values.TEST_HELPER.delete_test_data()

    # Lookup unixtime_s for a non-existent test with no data in data table.
    config['meta_data']['test_meta']['test_name'] = Values.TEST_HELPER.generate_random_string()
    loader = Loader(config)
    latest_unixtime_s = loader._Loader__lookup_latest_unixtime()
    assert (latest_unixtime_s is None)


@pytest.mark.database
@pytest.mark.load
def test_lookup_customer_id():

    config = deepcopy(Values.CONFIG_1)

    # Lookup existing customer id
    loader = Loader(config)
    customer_id = loader._Loader__lookup_customer_id()
    assert (customer_id == Values.TEST_HELPER.customer_id)
    del loader

    # Lookup non-existent customer.
    config['meta_data']['customers']['customer_name'] = Values.TEST_HELPER.generate_random_string()
    loader = Loader(config)
    customer_id = loader._Loader__lookup_customer_id()
    assert (customer_id is None)
    del loader


@pytest.mark.database
@pytest.mark.load
def test_lookup_project_id():

    config = deepcopy(Values.CONFIG_1)

    # Lookup existing project id
    loader = Loader(config)
    project_id = loader._Loader__lookup_project_id()
    assert (project_id == Values.TEST_HELPER.project_id)
    del loader

    # Lookup non-existent project.
    config['meta_data']['projects']['project_name'] = Values.TEST_HELPER.generate_random_string()
    loader = Loader(config)
    project_id = loader._Loader__lookup_project_id()
    assert (project_id is None)
    del loader


@pytest.mark.database
@pytest.mark.load
def test_lookup_latest_cycle():

    config = deepcopy(Values.CONFIG_1)

    Values.TEST_HELPER.delete_test_data()

    # Lookup latest cycle for a test that does not have data.
    loader = Loader(config)
    latest_cycle = loader._Loader__lookup_latest_cycle()
    assert (latest_cycle is None)

    # Define data to upload
    data_dict = {
        'test_id': [Values.TEST_HELPER.test_id, Values.TEST_HELPER.test_id],
        'cycle': [1, 2],
        'reported_charge_capacity_mah': [3000, 2999],
        'reported_charge_capacity_mah': [2995, 2994],
        'reported_coulombic_efficiency': [0.99, 0.98]
    }
    df = pd.DataFrame(data=data_dict)

    # Load data to the table.
    num_rows_inserted = Values.TEST_HELPER.load_df_to_db(
        df=df, target_table='test_data_cycle_stats')
    assert (num_rows_inserted == 2)

    # Check latest unixtime_s matches loaded data.
    latest_cycle = loader._Loader__lookup_latest_cycle()
    assert (latest_cycle == data_dict['cycle'][-1])

    Values.TEST_HELPER.delete_test_data()
    del loader

    # Lookup cycle for a test that does not exist.
    config['meta_data']['test_meta']['test_name'] = Values.TEST_HELPER.generate_random_string()
    loader = Loader(config)
    latest_cycle = loader._Loader__lookup_latest_cycle()
    assert (latest_cycle is None)


@pytest.mark.database
@pytest.mark.load
def test_load_test_data_small_dataset():

    config = deepcopy(Values.CONFIG_1)

    Values.TEST_HELPER.delete_test_data()

    # Define data to load.
    data_dict = {
        'cycle': [1, 1],
        'step': [1, 1],
        'test_time_s': [1.0, 2.0],
        'step_time_s': [1.0, 2.0],
        'current_ma': [0.0, 0.0],
        'voltage_mv': [3.789, 3.800],
        'recorded_datetime': [pd.Timestamp(1674659265, unit='s', tz='US/Pacific'),
                              pd.Timestamp(1674659266, unit='s', tz='US/Pacific')],
        'unixtime_s': [1674659265.0, 1674659266.0],
        'other_1': ['a', 'b'],
        'other_2': [1, 2]
    }
    df = pd.DataFrame(data=data_dict)

    # Load data
    loader = Loader(config)
    num_loaded_rows = loader.load_test_data(df)
    assert (num_loaded_rows == 2)

    # Pull loaded data and confirm it matches what we loaded.
    df_sql = pd.read_sql('SELECT * FROM test_data WHERE test_id = ' +
                         str(Values.TEST_HELPER.test_id) + ' ORDER BY unixtime_s;', Values.TEST_HELPER.engine)
    sql_dict = df_sql.to_dict(orient='list')
    # Convert the other_details column to a list of dicts.
    pd_other_details = [json.loads(row) for row in [
        '{"other_1": "a", "other_2": 1}',
        '{"other_1": "b", "other_2": 2}'
    ]]

    for key in data_dict.keys():
        if key == 'other_1' or key == 'other_2':
            # The other_1 and other_2 columns are stored as JSON strings.
            sq_dict_other_details = [row for row in sql_dict['other_details']]
            assert (sq_dict_other_details == pd_other_details)
        else:
            assert (data_dict[key] == sql_dict[key])

    # Try to re-load data that already exists in the database.
    num_loaded_rows = loader.load_test_data(df)
    assert (num_loaded_rows == 0)

    # Define more data to insert and load it
    new_data_dict = {
        'cycle': [1, 1],
        'step': [2, 2],
        'test_time_s': [3.0, 4.0],
        'step_time_s': [1.0, 2.0],
        'current_ma': [2.1, 2.1],
        'voltage_mv': [4.000, 4.020],
        'recorded_datetime': [pd.Timestamp(1674659267, unit='s', tz='US/Pacific'),
                              pd.Timestamp(1674659268, unit='s', tz='US/Pacific')],
        'unixtime_s': [1674659267.0, 1674659268.0]
    }
    df_new = pd.DataFrame(data=new_data_dict)
    num_loaded_rows = loader.load_test_data(df_new)
    assert (num_loaded_rows == 2)

    # Pull all loaded data and concat in to one dataset.
    df_sql = pd.read_sql('SELECT * FROM test_data WHERE test_id = ' +
                         str(Values.TEST_HELPER.test_id)+';', Values.TEST_HELPER.engine)
    sql_dict = df_sql.to_dict(orient='list')
    df = pd.concat([df, df_new])
    data_dict_combined = df.to_dict(orient='list')
    # Convert the other_details column to a list of dicts.
    pd_other_details_combined = [json.loads(row) if row else None for row in [
        '{"other_1": "a", "other_2": 1}',
        '{"other_1": "b", "other_2": 2}',
        None,
        None,
    ]]

    # Verify loaded data against all reference data.
    for key in data_dict_combined.keys():
        if key == 'other_1' or key == 'other_2':
            sq_dict_other_details = [
                row if row else None for row in sql_dict['other_details']]
            assert (pd_other_details_combined == pd_other_details_combined)
        else:
            assert (data_dict_combined[key] == sql_dict[key])

    # Delete the inserted data from database
    Values.TEST_HELPER.delete_test_data()


@pytest.mark.database
@pytest.mark.load
def test_load_cycle_stats_small_dataset():

    config = deepcopy(Values.CONFIG_1)

    Values.TEST_HELPER.delete_test_data()

    # Define data to load
    data_dict = {
        'cycle': [1, 2],
        'reported_charge_capacity_mah': [3000, 2999],
        'reported_discharge_capacity_mah': [2995, 2994],
        'reported_coulombic_efficiency': [0.99, 0.98]
    }
    df = pd.DataFrame(data=data_dict)

    # Load data
    loader = Loader(config)
    num_loaded_rows = loader.load_cycle_stats(df)
    assert (num_loaded_rows == 2)

    # Pull loaded data and verify it matches.
    df_sql = pd.read_sql(
        'SELECT * FROM test_data_cycle_stats WHERE test_id = '+str(Values.TEST_HELPER.test_id)+' ORDER BY cycle;', Values.TEST_HELPER.engine)
    sql_dict = df_sql.to_dict(orient='list')
    for key in data_dict.keys():
        assert (data_dict[key] == sql_dict[key])

    # Overwrite the existing data with slightly different data.
    data_dict = {
        'cycle': [1, 2],
        'reported_charge_capacity_mah': [3001, 2998],
        'reported_discharge_capacity_mah': [2996, 2995],
        'reported_coulombic_efficiency': [0.95, 0.96]
    }
    df = pd.DataFrame(data=data_dict)
    loader = Loader(config)
    num_loaded_rows = loader.load_cycle_stats(df)
    assert (num_loaded_rows == 2)

    # Pull loaded data and verify it matches.
    df_sql = pd.read_sql(
        'SELECT * FROM test_data_cycle_stats WHERE test_id = '+str(Values.TEST_HELPER.test_id)+' ORDER BY cycle;', Values.TEST_HELPER.engine)
    sql_dict = df_sql.to_dict(orient='list')
    for key in data_dict.keys():
        assert (data_dict[key] == sql_dict[key])

    # Define more data to insert and load it
    new_data_dict = {
        'cycle': [3, 4],
        'reported_charge_capacity_mah': [2998, 2997],
        'reported_discharge_capacity_mah': [2993, 2992],
        'reported_coulombic_efficiency': [0.97, 0.97]
    }
    df_new = pd.DataFrame(data=new_data_dict)
    num_loaded_rows = loader.load_cycle_stats(df_new)
    assert (num_loaded_rows == 2)

    # Pull all loaded data and verify
    df_sql = pd.read_sql(
        'SELECT * FROM test_data_cycle_stats WHERE test_id = '+str(Values.TEST_HELPER.test_id)+' ORDER BY cycle;', Values.TEST_HELPER.engine)
    sql_dict = df_sql.to_dict(orient='list')
    df = pd.concat([df, df_new])
    data_dict_combined = df.to_dict(orient='list')
    for key in data_dict_combined.keys():
        assert (data_dict_combined[key] == sql_dict[key])

    Values.TEST_HELPER.delete_test_data()


@pytest.mark.database
@pytest.mark.load
def test_insert_cycler_meta():

    config = deepcopy(Values.CONFIG_1)

    # Load new cycler meta
    config['meta_data']['cycler_meta']['model'] = Values.TEST_HELPER.generate_random_string()
    loader = Loader(config)
    cycler_type_id = loader._Loader__insert_cycler_meta()
    assert (cycler_type_id)

    # Make sure loaded data matches what is in config.
    new_row = Values.TEST_HELPER.read_last_row(
        target_table='cyclers_meta', pk_col_name='cycler_type_id')
    new_cycler_type_id = new_row[0]
    new_manufacturer = new_row[1]
    new_model = new_row[2]
    assert (new_cycler_type_id == cycler_type_id)
    assert (new_manufacturer ==
            config['meta_data']['cycler_meta']['manufacturer'])
    assert (new_model == config['meta_data']['cycler_meta']['model'])

    Values.TEST_HELPER.delete_entry(target_table='cyclers_meta',
                                    pk_col_name='cycler_type_id', pk_id=cycler_type_id)


@pytest.mark.database
@pytest.mark.load
def test_insert_cycler():

    config = deepcopy(Values.CONFIG_1)

    # Test inserting a cycler with a cycler_type_id that already exists.
    config['meta_data']['cycler']['sn'] = Values.TEST_HELPER.generate_random_string()
    loader = Loader(config)
    cycler_type_id = loader._Loader__lookup_cycler_type_id()
    assert (cycler_type_id)
    cycler_id = loader._insert_cycler()
    assert (cycler_id)

    # Make sure loaded data matches what is in config.
    new_row = Values.TEST_HELPER.read_last_row(
        target_table='cyclers', pk_col_name='cycler_id')
    newest_cycler_id = new_row[0]
    newest_cycler_type_id = new_row[1]
    newest_sn = new_row[2]
    assert (newest_cycler_id == cycler_id)
    assert (newest_cycler_type_id == cycler_type_id)
    assert (newest_sn == config['meta_data']['cycler']['sn'])

    del loader
    Values.TEST_HELPER.delete_entry(target_table='cyclers',
                                    pk_col_name='cycler_id', pk_id=cycler_id)

    # Test inserting a cycler with with a new cycler_type.
    config['meta_data']['cycler_meta']['manufacturer'] = Values.TEST_HELPER.generate_random_string()
    config['meta_data']['cycler_meta']['model'] = Values.TEST_HELPER.generate_random_string()
    config['meta_data']['cycler']['sn'] = Values.TEST_HELPER.generate_random_string()
    loader = Loader(config)
    cycler_type_id = loader._Loader__lookup_cycler_type_id()
    assert (cycler_type_id is None)
    cycler_id = loader._insert_cycler()
    assert (cycler_id)
    cycler_type_id = loader._Loader__lookup_cycler_type_id()
    assert (cycler_type_id)

    # Get the latest entry in the cyclers table.
    new_row = Values.TEST_HELPER.read_last_row(
        target_table='cyclers', pk_col_name='cycler_id')
    newest_cycler_id = new_row[0]
    newest_cycler_type_id = new_row[1]
    newest_sn = new_row[2]
    assert (newest_cycler_id == cycler_id)
    assert (newest_cycler_type_id == cycler_type_id)
    assert (newest_sn == config['meta_data']['cycler']['sn'])

    Values.TEST_HELPER.delete_entry(target_table='cyclers',
                                    pk_col_name='cycler_id', pk_id=cycler_id)
    Values.TEST_HELPER.delete_entry(target_table='cyclers_meta',
                                    pk_col_name='cycler_type_id', pk_id=cycler_type_id)


@pytest.mark.database
@pytest.mark.load
def test_insert_cell_meta():

    config = deepcopy(Values.CONFIG_1)

    # Insert non-existing cell_meta.
    config['meta_data']['cell_meta']['manufacturer'] = Values.TEST_HELPER.generate_random_string()
    config['meta_data']['cell_meta']['manufacturer_pn'] = Values.TEST_HELPER.generate_random_string()
    loader = Loader(config)
    cell_type_id = loader._Loader__lookup_cell_type_id()
    assert (cell_type_id is None)
    cell_type_id = loader._Loader__insert_cell_meta()
    assert (cell_type_id)

    new_row = Values.TEST_HELPER.read_last_row(
        target_table='cells_meta', pk_col_name='cell_type_id')
    newest_cell_type_id = new_row[0]
    newest_manufacturer = new_row[1]
    newest_manufacturer_pn = new_row[2]
    assert (newest_cell_type_id == cell_type_id)
    assert (newest_manufacturer ==
            config['meta_data']['cell_meta']['manufacturer'])
    assert (newest_manufacturer_pn ==
            config['meta_data']['cell_meta']['manufacturer_pn'])

    del loader
    Values.TEST_HELPER.delete_entry(target_table='cells_meta',
                                    pk_col_name='cell_type_id', pk_id=cell_type_id)


@pytest.mark.database
@pytest.mark.load
def test_insert_cell():

    config = deepcopy(Values.CONFIG_1)

    # Test inserting a cell with a cell_type_id that already exists.
    config['meta_data']['cell']['manufacturer_sn'] = Values.TEST_HELPER.generate_random_string()
    config['meta_data']['cell']['label'] = Values.TEST_HELPER.generate_random_string()
    loader = Loader(config)
    cell_type_id = loader._Loader__lookup_cell_type_id()
    assert (cell_type_id)
    cell_id = loader._Loader__insert_cell()
    assert (cell_id)

    new_row = Values.TEST_HELPER.read_last_row(
        target_table='cells', pk_col_name='cell_id')
    newest_cell_id = new_row[0]
    newest_cell_type_id = new_row[1]
    newest_sn = new_row[7]
    newest_label = new_row[3]
    assert (newest_cell_id == cell_id)
    assert (newest_cell_type_id == cell_type_id)
    assert (newest_sn == config['meta_data']['cell']['manufacturer_sn'])
    assert(newest_label == config['meta_data']['cell']['label'])

    del loader
    Values.TEST_HELPER.delete_entry(target_table='cells',
                                    pk_col_name='cell_id', pk_id=cell_id)

    # Test inserting a cycler with a cycler_type_id that does not exist.
    config['meta_data']['cell_meta']['manufacturer'] = Values.TEST_HELPER.generate_random_string()
    config['meta_data']['cell_meta']['manufacturer_pn'] = Values.TEST_HELPER.generate_random_string()
    config['meta_data']['cell']['manufacturer_sn'] = Values.TEST_HELPER.generate_random_string()
    loader = Loader(config)
    cell_type_id = loader._Loader__lookup_cell_type_id()
    assert (cell_type_id is None)
    cell_id = loader._Loader__insert_cell()
    assert (cell_id)
    cell_type_id = loader._Loader__lookup_cell_type_id()
    assert (cell_type_id)

    new_row = Values.TEST_HELPER.read_last_row(
        target_table='cells', pk_col_name='cell_id')
    newest_cell_id = new_row[0]
    newest_cell_type_id = new_row[1]
    newest_sn = new_row[7]
    assert (newest_cell_id == cell_id)
    assert (newest_cell_type_id == cell_type_id)
    assert (newest_sn == config['meta_data']['cell']['manufacturer_sn'])

    del loader
    Values.TEST_HELPER.delete_entry(target_table='cells',
                                    pk_col_name='cell_id', pk_id=cell_id)
    Values.TEST_HELPER.delete_entry(target_table='cells_meta',
                                    pk_col_name='cell_type_id', pk_id=cell_type_id)


@pytest.mark.database
@pytest.mark.load
def test_insert_schedule_meta():

    config = deepcopy(Values.CONFIG_1)

    # Insert a new schedule
    config['meta_data']['schedule_meta']['schedule_name'] = Values.TEST_HELPER.generate_random_string()
    loader = Loader(config)
    schedule_id = loader._Loader__lookup_schedule_id()
    assert (schedule_id is None)
    schedule_id = loader._Loader__insert_schedule_meta()
    assert (schedule_id)

    new_row = Values.TEST_HELPER.read_last_row(
        target_table='schedule_meta', pk_col_name='schedule_id')
    newest_schedule_id = new_row[0]
    newest_schedule_file_name = new_row[1]
    assert (newest_schedule_id == schedule_id)
    assert (newest_schedule_file_name ==
            config['meta_data']['schedule_meta']['schedule_name'])

    Values.TEST_HELPER.delete_entry(
        target_table='schedule_meta', pk_col_name='schedule_id', pk_id=schedule_id)


@pytest.mark.database
@pytest.mark.load
def test_insert_test_meta():

    config = deepcopy(Values.CONFIG_1)

    # Test inserting a test with a cell, schedule, and cycler that already exist
    config['meta_data']['test_meta']['test_name'] = Values.TEST_HELPER.generate_random_string()
    loader = Loader(config)
    cell_id = loader._lookup_cell_id()
    assert (cell_id)
    schedule_id = loader._Loader__lookup_schedule_id()
    assert (schedule_id)
    cycler_id = loader._lookup_cycler_id()
    assert (cycler_id)
    test_id = loader._Loader__insert_test_meta()
    assert (test_id)

    new_row = Values.TEST_HELPER.read_last_row(
        target_table='test_meta', pk_col_name='test_id')
    newest_test_id = new_row[0]
    newest_cell_id = new_row[1]
    newest_schedule_id = new_row[2]
    newest_test_name = new_row[3]
    newest_cycler_id = new_row[15]
    assert (newest_test_id == test_id)
    assert (newest_cell_id == cell_id)
    assert (newest_cycler_id == cycler_id)
    assert (newest_schedule_id == schedule_id)
    assert (newest_test_name == config['meta_data']['test_meta']['test_name'])

    del loader
    Values.TEST_HELPER.delete_entry(target_table='test_meta',
                                    pk_col_name='test_id', pk_id=test_id)

    # Now insert a test with new everything
    config = deepcopy(Values.CONFIG_1)
    config['meta_data']['test_meta']['test_name'] = Values.TEST_HELPER.generate_random_string()
    config['meta_data']['schedule_meta']['schedule_name'] = Values.TEST_HELPER.generate_random_string()
    config['meta_data']['cell']['manufacturer_sn'] = Values.TEST_HELPER.generate_random_string()
    config['meta_data']['cell_meta']['manufacturer'] = Values.TEST_HELPER.generate_random_string()
    config['meta_data']['cell_meta']['manufacturer_pn'] = Values.TEST_HELPER.generate_random_string()
    config['meta_data']['cycler']['sn'] = Values.TEST_HELPER.generate_random_string()
    config['meta_data']['cycler_meta']['manufacturer'] = Values.TEST_HELPER.generate_random_string()
    config['meta_data']['cycler_meta']['model'] = Values.TEST_HELPER.generate_random_string()
    loader = Loader(config)

    # Make sure no fields existed before.
    assert (loader._lookup_cell_id() is None)
    assert (loader._Loader__lookup_cell_type_id() is None)
    assert (loader._Loader__lookup_schedule_id() is None)
    assert (loader._lookup_cycler_id() is None)
    assert (loader._Loader__lookup_cycler_type_id() is None)

    # Insert new test data
    test_id = loader._Loader__insert_test_meta()
    assert (test_id)

    # Get the IDs for the newly inserted data
    schedule_id = loader._Loader__lookup_schedule_id()
    assert (schedule_id)
    cycler_id = loader._lookup_cycler_id()
    assert (cycler_id)
    cycler_type_id = loader._Loader__lookup_cycler_type_id()
    assert (cycler_type_id)
    cell_id = loader._lookup_cell_id()
    assert (cell_id)
    cell_type_id = loader._Loader__lookup_cell_type_id()
    assert (cell_type_id)

    # Make sure all uploaded data matches
    new_row = Values.TEST_HELPER.read_last_row(
        target_table='test_meta', pk_col_name='test_id')
    newest_test_id = new_row[0]
    newest_cell_id = new_row[1]
    newest_schedule_id = new_row[2]
    newest_test_name = new_row[3]
    newest_cycler_id = new_row[15]
    assert (newest_test_id == test_id)
    assert (newest_cell_id == cell_id)
    assert (newest_cycler_id == cycler_id)
    assert (newest_schedule_id == schedule_id)
    assert (newest_test_name == config['meta_data']['test_meta']['test_name'])

    del loader
    Values.TEST_HELPER.delete_entry(
        target_table='test_meta', pk_col_name='test_id', pk_id=test_id)
    Values.TEST_HELPER.delete_entry(
        target_table='cyclers', pk_col_name='cycler_id', pk_id=cycler_id)
    Values.TEST_HELPER.delete_entry(
        target_table='cyclers_meta', pk_col_name='cycler_type_id', pk_id=cycler_type_id)
    Values.TEST_HELPER.delete_entry(
        target_table='cells', pk_col_name='cell_id', pk_id=cell_id)
    Values.TEST_HELPER.delete_entry(
        target_table='cells_meta', pk_col_name='cell_type_id', pk_id=cell_type_id)
    Values.TEST_HELPER.delete_entry(
        target_table='schedule_meta', pk_col_name='schedule_id', pk_id=schedule_id)
